
import math

import cv2
import numpy as np
from scipy.spatial.distance import cdist

from AM_CommonTools.data.connected_component import ConnectedComponent
from AM_CommonTools.util.geometry_helper import GeometryHelper
from AM_CommonTools.util.line_segment_2D import LineSegment2D
from AM_CommonTools.util.polygon_2D import Polygon2D


class SpatialCCGraph:
    DistanceBoxes = 0
    DistanceContours = 1
    DistanceCenters = 2
    PolygonBoundingBox = 0
    PolygonConvexHull = 1
    WildcardLabel = "<WILDCARD>"

    def __init__(self, initial_vertices, width=None, height=None):
        for cc in initial_vertices:
            assert isinstance(cc, ConnectedComponent)

        self.vertices = initial_vertices
        self.vertices_by_id = {cc.strID(): idx for idx, cc in enumerate(self.vertices)}

        # connection exist (yes/no)
        self.connections = np.zeros((len(self.vertices), len(self.vertices)), dtype=np.bool)

        # vertices top classes
        self.vertices_labels = None
        self.vertices_confidences = None
        self.max_prod_pairs_weight = None
        self.max_comb_pairs_weight = None
        self.cache_3D_vectors = None

        # original image dimensions
        self.width = width
        self.height = height

    def optimize_for_storage(self):
        if self.vertices is None:
            return

        for cc in self.vertices:
            # drop CC image reference ...
            cc.img = None

        self.cache_3D_vectors = None
        self.vertices_by_id = None

    def set_vertices_classes(self, vertices_classes):
        self.vertices_labels = []
        self.vertices_confidences = []

        # for each vertex ...
        for vertex_classes in vertices_classes:
            vertex_labels = []
            vertex_confidences = []

            # for each class along with its confidence ...
            for label, confidence in vertex_classes:
                # assume labels are sorted by desc. confidence. Stop when 0 has been found
                if confidence == 0.0:
                    break

                vertex_labels.append(label)
                vertex_confidences.append(confidence)

            self.vertices_labels.append(vertex_labels)
            self.vertices_confidences.append(vertex_confidences)

    def recover_for_rendering(self, original_ccs):
        original_per_id = {}
        pending_on_original = []
        for cc in original_ccs:
            cc_id = cc.strID()
            original_per_id[cc_id] = cc
            pending_on_original.append(cc_id)

        pending_on_current = []
        for cc in self.vertices:
            cc_id = cc.strID()
            if cc_id in original_per_id:
                cc.img = original_per_id[cc.strID()].img
            else:
                pending_on_current.append(cc_id)

        # try to find original CCs for each pending on current ...
        # (This only works based on the assumption that these CCs are the product of merging two CC's from the original)
        for local_cc_id in pending_on_current:
            local_cc = self.vertices[self.vertices_by_id[local_cc_id]]
            # identify contained candidates from pending elements of the original ...
            merge_candidates = []
            for original_cc_id in pending_on_original:
                original_cc = original_per_id[original_cc_id]

                if (local_cc.min_x <= original_cc.min_x and original_cc.max_x <= local_cc.max_x and
                    local_cc.min_y <= original_cc.min_y and original_cc.max_y <= local_cc.max_y):
                    merge_candidates.append(original_cc_id)

            for idx_1 in range(len(merge_candidates) - 1):
                merge_cc_1 = original_per_id[merge_candidates[idx_1]]
                for idx_2 in range(1, len(merge_candidates)):
                    merge_cc_2 = original_per_id[merge_candidates[idx_2]]

                    if (local_cc.min_x == min(merge_cc_1.min_x, merge_cc_2.min_x) and
                        local_cc.max_x == max(merge_cc_1.max_x, merge_cc_2.max_x) and
                        local_cc.min_y == min(merge_cc_1.min_y, merge_cc_2.min_y) and
                        local_cc.max_y == max(merge_cc_1.max_y, merge_cc_2.max_y)):
                        # found
                        local_cc.img = ConnectedComponent.Merge([merge_cc_1, merge_cc_2]).img
                        break

                if local_cc.img is not None:
                    break

            if local_cc.img is None:
                raise Exception("Could not find the original CCs of " + str(local_cc))

    def approximate_recover_for_rendering(self, approximate_ccs):
        # similar to recover_for_rendering, but this version assumes that bounding boxes will not be identical
        # and that no explicit merges have been performed on source, but still some CCs in the approximated image
        # can belong to multiple CCs in the spatial graph ...

        # small to large?
        # or greedy by overlap ratio ( intersection / union )

        # greedily find matches by Intersection Over Union
        match_candidates = []
        local_matches_by_id = {}
        for local_cc in self.vertices:
            local_cc_id = local_cc.strID()
            local_matches_by_id[local_cc_id] = []

            for other_idx, other_cc in enumerate(approximate_ccs):
                other_cc_id = other_cc.strID()

                if ((local_cc.min_x <= other_cc.max_x and other_cc.min_x <= local_cc.max_x) and
                    (local_cc.min_y <= other_cc.max_y and other_cc.min_y <= local_cc.max_y)):
                    # compute intersection ...
                    i_min_x = max(local_cc.min_x, other_cc.min_x)
                    i_max_x = min(local_cc.max_x, other_cc.max_x)
                    i_min_y = max(local_cc.min_y, other_cc.min_y)
                    i_max_y = min(local_cc.max_y, other_cc.max_y)

                    int_area = (i_max_x - i_min_x + 1) * (i_max_y - i_min_y + 1)

                    # compute union area ...
                    union_area = local_cc.getBoxArea() + other_cc.getBoxArea() - int_area

                    IoU = int_area / union_area
                    match_candidates.append((IoU, local_cc_id, other_cc_id, other_idx))

        # sort by descending intersection/union
        match_candidates = sorted(match_candidates, reverse=True)

        other_matches_by_id = {}
        for IoU, local_cc_id, other_cc_id, other_idx in match_candidates:
            # force 1-local to N-other
            # (CCs from original can belong only to one local
            #  but local can have multiple from other)
            if not other_cc_id in other_matches_by_id:
                # mark as matched ...
                other_matches_by_id[other_cc_id] = local_cc_id

                # append to local matches ...
                local_matches_by_id[local_cc_id].append(other_idx)

        for local_cc_id in local_matches_by_id:
            local_cc = self.vertices[self.vertices_by_id[local_cc_id]]
            w = local_cc.getWidth()
            h = local_cc.getHeight()

            if len(local_matches_by_id[local_cc_id]) == 0:
                # print(match_candidates)
                # print(local_matches_by_id)
                # print(local_cc_id)
                # print((self.width, self.height))
                print("Warning: Could not find matches for CC", flush=True)
                # create empty image ...
                img = np.zeros((h, w), np.uint8)
                # middle pixel is white?
                img[int(h / 2), int(w / 2)] = 255
                local_cc.img = img
            else:
                if len(local_matches_by_id[local_cc_id]) > 1:
                    # merge ...
                    merge_list = [approximate_ccs[other_idx] for other_idx in local_matches_by_id[local_cc_id]]
                    base_img = ConnectedComponent.Merge(merge_list).img
                else:
                    base_img = approximate_ccs[local_matches_by_id[local_cc_id][0]].img

                # resize ...
                if base_img.shape[0] != h or base_img.shape[1] != w:
                    base_img = cv2.resize(base_img, (w, h))

                # print((w, h, str(local_cc), base_img.shape, self.width, self.height))
                local_cc.img = base_img


    def setDimensions(self, width, height):
        self.width = width
        self.height = height

    def selectVerticesByRegion(self, min_x, max_x, min_y, max_y):
        selected_idxs = []
        for idx, cc in enumerate(self.vertices):
            if (cc.min_x < max_x and min_x < cc.max_x) and (cc.min_y < max_y and min_y < cc.max_y):
                selected_idxs.append(idx)

        return selected_idxs

    def makeFullyConnected(self):
        # mark everything as connected ...
        self.connections[:, :] = True

    def is_small_graph(self, min_graph_size):
        return min_graph_size > 0 and len(self.vertices) <= min_graph_size

    def add_self_connections(self):
        for idx in range(self.connections.shape[0]):
            self.connections[idx, idx] = True

    def vertexContourDistances(self):
        # get all contour points from all CC
        contour_points = [np.vstack(cc.getContours()) for cc in self.vertices]

        # initialize matrix
        n_vertices = len(self.vertices)
        distance_matrix = np.zeros((n_vertices, n_vertices), dtype=np.float64)

        for idx1 in range(n_vertices - 1):
            for idx2 in range(idx1 + 1, n_vertices):
                # compute distance between all contour points ...
                pair_matrix = cdist(contour_points[idx1], contour_points[idx2])
                distance = pair_matrix.min()

                distance_matrix[idx1, idx2] = distance
                distance_matrix[idx2, idx1] = distance

        return distance_matrix

    def vertexBoxDistances(self):
        # initialize matrix
        n_vertices = len(self.vertices)
        distance_matrix = np.zeros((n_vertices, n_vertices), dtype=np.float64)

        for idx1 in range(n_vertices - 1):
            for idx2 in range(idx1 + 1, n_vertices):
                # get distance between bounding boxes ...
                distance = self.vertices[idx1].getBoxDistance(self.vertices[idx2])

                distance_matrix[idx1, idx2] = distance
                distance_matrix[idx2, idx1] = distance

        return distance_matrix

    def vertexCenterDistances(self):
        n_vertices = len(self.vertices)
        distance_matrix = np.zeros((n_vertices, n_vertices), dtype=np.float64)

        for idx1 in range(n_vertices - 1):
            for idx2 in range(idx1 + 1, n_vertices):
                # get distance between bounding boxes ...
                distance = self.vertices[idx1].getCenterDistance(self.vertices[idx2])

                distance_matrix[idx1, idx2] = distance
                distance_matrix[idx2, idx1] = distance

        return distance_matrix

    def vertexDistances(self, distance_type):
        # compute distances between CCs
        if distance_type == SpatialCCGraph.DistanceContours:
            return self.vertexContourDistances()
        elif distance_type == SpatialCCGraph.DistanceBoxes:
            return self.vertexBoxDistances()
        elif distance_type == SpatialCCGraph.DistanceCenters:
            return self.vertexCenterDistances()
        else:
            raise Exception("Unknown Distance type for vertices")

    def __findPotentialVertexOccluders(self, i, j, polygons):
        # find vertices that could potentially block the sight between vertices i and j
        pair_box = GeometryHelper.boxesBoudingBox([self.vertices[i].getBoundingBox(),
                                                   self.vertices[j].getBoundingBox()])

        # find potential occlusions ...
        potential_occluders = []
        for k in range(len(self.vertices)):
            if k != i and k != j and GeometryHelper.boxesIntercept(pair_box, self.vertices[k].getBoundingBox()):
                potential_occluders.append(polygons[k])

        return potential_occluders

    def __polygonInLineOfSight(self, vertex_i, target_polygon, potential_occluders):
        # check the line between the eye of i and every point in the target polygon
        # defines a eye at the center of the vertex...
        eye_x, eye_y = vertex_i.getCenter()

        for target_x, target_y in target_polygon.points:
            line_of_sight = LineSegment2D((eye_x, eye_y), (target_x, target_y))

            # check for occlussions of LOS with third objects....
            occlusion_found = False

            for polygon in potential_occluders:
                occlusion_found = polygon.lineIntersects(line_of_sight)
                if occlusion_found:
                    break

            # check if i can see j using current LOS
            if not occlusion_found:
                # mark as visible ...
                return True

        # no line of sight found ...
        return False

    def makeLOS(self, polygon_type, max_distance=None, distance_type=None):
        if max_distance is not None and distance_type is None:
            raise Exception("Must specify distance type")

        if distance_type is not None:
            vertex_distances = self.vertexDistances(distance_type)
        else:
            vertex_distances = None

        # get a representation of boundaries for each vertex
        polygons = []
        for idx, cc in enumerate(self.vertices):
            if polygon_type == SpatialCCGraph.PolygonConvexHull:
                contour_points = np.vstack(cc.getContours())

                convex_hull = cv2.convexHull(contour_points)
                convex_hull = convex_hull.reshape((convex_hull.shape[0], convex_hull.shape[2]))

                polygons.append(Polygon2D(convex_hull.tolist()))

            elif polygon_type == SpatialCCGraph.PolygonBoundingBox:
                bounding_box = [[cc.min_x, cc.min_y], [cc.max_x, cc.min_y], [cc.max_x, cc.max_y], [cc.min_x, cc.max_y]]

                polygons.append(Polygon2D(bounding_box))

        # now, compute the LOS algorithm....
        n_vertices = len(self.vertices)
        cc_visible = np.zeros((n_vertices, n_vertices), dtype=np.bool)

        # for each vertex i....
        for i in range(n_vertices - 1):
            # check every vertex j....
            for j in range(i + 1, n_vertices):
                # check if i can see j
                # (note that i could not directly see j from it's eye, but j's eye might see i)

                # ... first, if max distance is specified ...
                if max_distance is not None and vertex_distances[i, j] > max_distance:
                    # too far way, do not check for connections ...
                    continue

                # get the subset of vertices that are potential occluders ...
                potential_occluders = self.__findPotentialVertexOccluders(i, j, polygons)

                if (self.__polygonInLineOfSight(self.vertices[i], polygons[j], potential_occluders) or
                    self.__polygonInLineOfSight(self.vertices[j], polygons[i], potential_occluders)):
                    # i can see j and/or j can see i ...
                    # mark them as mutually visible....
                    cc_visible[i, j] = True
                    cc_visible[j, i] = True

        # Now, replace connections
        self.connections = cc_visible

    def makeKNN(self, K, distance_type):
        # compute distances between CCs
        vertex_distances = self.vertexDistances(distance_type)

        # create connections
        knn_conn = np.zeros((len(self.vertices), len(self.vertices)), dtype=np.bool)

        for v in range(len(self.vertices)):
            # sort neighbors by distance
            v_distances = sorted([(value, u) for u, value in enumerate(vertex_distances[v, :])])
            # remove itself ...
            v_distances.remove((0.0, v))

            # for the top-K closest neighbors
            for distance, u in v_distances[:K]:
                knn_conn[v, u] = True
                knn_conn[u, v] = True

        self.connections = knn_conn

    def makePseudoMST(self, distance_type):
        # use Kruskal's algorithm
        # first, generate a list of edges sorted by cost...
        edge_list = []
        previous = []

        vertex_distances = self.vertexDistances(distance_type)

        #cdist(np.vstack(z.nonzero()).T, y)
        for v in range(len(self.vertices)):
            for u in range(v):
                cost = vertex_distances[v, u]
                edge_list.append((cost, v, u))

        # ...sort...
        edge_list = sorted( edge_list )

        # ...empty graph...
        mst_conn = np.zeros((len(self.vertices), len(self.vertices)), dtype=np.bool)

        # ... create the forest, all nodes are disconnected and belong to
        # ... different sets...
        sets = {u : u for u in range(len(self.vertices))}

        # now, for each edge of minimum cost, add it if it does not form a cycle...
        joins = 0
        for cost, v, u in edge_list:
            # check...
            if sets[u] != sets[v]:
                # belong to different sets...
                joins += 1

                # now, make the union of the two sets...
                # add all elements of the set of v to the set of u
                v_set = sets[v]
                for w in range(len(self.vertices)):
                    if sets[w] == v_set:
                        sets[w] = sets[u]

                # add the edge to the graph...
                mst_conn[u, v] = True
                mst_conn[v, u] = True

            # check if finished...
            if joins >= len(self.vertices) - 1:
                break

        self.connections = mst_conn

    def set_connection(self, cc_str_id1, cc_str_id2, connect):
        idx1 = self.vertices_by_id[cc_str_id1]
        idx2 = self.vertices_by_id[cc_str_id2]

        self.connections[idx1, idx2] = connect
        self.connections[idx2, idx1] = connect

    def disconnect_node(self, cc_str_id):
        current = self.vertices_by_id[cc_str_id]

        for other in range(len(self.vertices)):
            self.connections[current, other] = False
            self.connections[other, current] = False

    def add(self, new_cc):
        # add to vertices ...
        self.vertices.append(new_cc)

        # expand connection matrix ...
        exp_conn = np.zeros((len(self.vertices), len(self.vertices)), dtype=np.bool)
        exp_conn[:-1, :-1] = self.connections.copy()
        self.connections = exp_conn

        self.update_index()

    def delete(self, cc_str_id):
        if cc_str_id in self.vertices_by_id:
            idx = self.vertices_by_id[cc_str_id]

            del self.vertices[idx]

            red_conn = np.zeros((len(self.vertices), len(self.vertices)), dtype=np.bool)
            # copy other connections ...
            # top-left
            red_conn[:idx, :idx] = self.connections[:idx, :idx].copy()
            # top-right
            red_conn[:idx, idx:] = self.connections[:idx, idx + 1:].copy()
            # bottom-left
            red_conn[idx:, :idx] = self.connections[idx + 1:, :idx].copy()
            # bottom-right
            red_conn[idx:, idx:] = self.connections[idx + 1:, idx + 1:].copy()

            self.connections = red_conn

            self.update_index()

    def merge_vertices(self, cc_str_id_1, cc_str_id_2):
        # create combined image (probably no longer a real CC)...
        idx1 = self.vertices_by_id[cc_str_id_1]
        idx2 = self.vertices_by_id[cc_str_id_2]

        if idx2 < idx1:
            # swap indices and string ids ...
            tempo_idx = idx1
            idx1 = idx2
            idx2 = tempo_idx

            tempo_str_id = cc_str_id_1
            cc_str_id_1 = cc_str_id_2
            cc_str_id_2 = tempo_str_id

        combined_cc = ConnectedComponent.Merge([self.vertices[idx1], self.vertices[idx2]])

        # combine connections by adding connections of CC #2 to connections of CC #1
        for other_idx in range(len(self.vertices)):
            conn_value = (self.connections[idx1, other_idx] or self.connections[idx2, other_idx])
            self.connections[idx1, other_idx] = conn_value
            self.connections[other_idx, idx1] = conn_value

        # delete second CC
        self.delete(cc_str_id_2)

        # replace CC at 1
        self.vertices[idx1] = combined_cc

        # re-index ...
        self.update_index()

        return combined_cc

    def update_index(self):
        self.vertices_by_id = {cc.strID(): idx for idx, cc in enumerate(self.vertices)}

    def __image_render_cc(self, image, cc, color):
        assert isinstance(cc, ConnectedComponent)

        cc_cut = image[cc.min_y:cc.max_y + 1, cc.min_x:cc.max_x + 1, :]
        cc_cut[cc.img > 0, 0] = color[0]
        cc_cut[cc.img > 0, 1] = color[1]
        cc_cut[cc.img > 0, 2] = color[2]

    def get_connections(self):
        duplicated_connections = np.vstack(self.connections.nonzero()).T

        # upper row + main diagonal
        upper_row = duplicated_connections[:, 0] <= duplicated_connections[:, 1]

        return duplicated_connections[upper_row]

    def total_connections(self):
        # consider unique connections only
        # (a link from i to j and from j to i counts as one link only)
        total_sum = self.connections.sum()

        # but the main diagonal has been counted only once
        for idx in range(self.connections.shape[0]):
            # add these elements again ...
            total_sum += self.connections[idx, idx]

        return int(total_sum / 2)

    def render(self, highlight_vertices=None, highlight_connections=None, context=None, style=None):
        if self.width is None or self.height is None:
            raise Exception("ERROR: Cannnot Render Graph! Dimensions not set!")

        result = np.ones((self.height, self.width, 3), np.uint8)

        if highlight_vertices is None:
            highlight_vertices = []
        if highlight_connections is None:
            highlight_connections = []

        # defaults ...
        color_bg = [255, 255, 255]
        color_cc_context = [192, 192, 192]
        color_cc_normal = [0, 0, 0]
        color_cc_highlight = [34, 177, 76]
        color_edge_normal = [0, 0, 64]
        color_edge_highlight = [255, 0, 0]
        color_multicolor_highlight = False
        draw_edges = True
        width_edge_normal = 1
        width_edge_highlight = 2
        multi_color_highlight = [
            [192, 192, 0], [192, 0, 192],  [0, 192, 192],
            [0, 128, 0], [128, 0, 0],  [0, 0, 128],
            [128, 128, 0], [128, 0, 128],  [0, 128, 128],
            [255, 128, 64], [255, 64, 128],
            [128, 255, 64], [128, 255, 64],
            [128, 64, 255], [64, 128, 255],
            [0, 128, 64], [0, 64, 128],
            [128, 0, 64], [128, 0, 64],
            [128, 64, 0], [64, 128, 0],
        ]

        # check for overriden ...
        if style is not None:
            if "bg_color" in style:
                color_bg = style["bg_color"]

            if "draw_edges" in style:
                draw_edges = style["draw_edges"]

            if "use_multicolor" in style:
                color_multicolor_highlight = style["use_multicolor"]

            if "cc_context_color" in style:
                color_cc_context = style["cc_context_color"]

            if "cc_normal_color" in style:
                color_cc_normal = style["cc_normal_color"]

            if "cc_highlight_color" in style:
                color_cc_highlight = style["cc_highlight_color"]

            if "edge_normal_color" in style:
                color_edge_normal = style["edge_normal_color"]

            if "edge_highlight_color" in style:
                color_edge_highlight = style["edge_highlight_color"]

            if "edge_normal_width" in style:
                width_edge_normal = style["edge_normal_width"]

            if "edge_highlight_width" in style:
                width_edge_highlight = style["edge_highlight_width"]

        # coloring the background ...
        result[:, :, 0] *= color_bg[0]
        result[:, :, 1] *= color_bg[1]
        result[:, :, 2] *= color_bg[2]

        # determine connections to be rendered ...
        lines_normal = []
        lines_highlight = []
        if draw_edges:
            for idx1 in range(len(self.vertices) - 1):
                cc1 = self.vertices[idx1]
                cx1, cy1 = cc1.getCenter()

                for idx2 in range(idx1 + 1, len(self.vertices)):
                    cc2 = self.vertices[idx2]
                    cx2, cy2 = cc2.getCenter()

                    if self.connections[idx1][idx2]:
                        e1 = (cc1.strID(), cc2.strID())
                        e2 = (cc2.strID(), cc1.strID())

                        if e1 in highlight_connections or e2 in highlight_connections:
                            current_color = color_edge_highlight
                            current_width = width_edge_highlight
                        else:
                            current_color = color_edge_normal
                            current_width = width_edge_normal

                        inv_color = (current_color[2], current_color[1], current_color[0])

                        line_info = ((int(cx1), int(cy1)), (int(cx2), int(cy2)), inv_color, current_width)
                        if e1 in highlight_connections or e2 in highlight_connections:
                            lines_highlight.append(line_info)
                        else:
                            lines_normal.append(line_info)

        # rendering extra context CC (if any)
        if context is not None:
            for cc in context:
                self.__image_render_cc(result, cc, color_cc_context)

        # render normal edges ...
        for p1, p2, color, width in lines_normal:
            cv2.line(result, p1, p2, color, thickness=width)

        # render the normal CCs ...
        for cc in self.vertices:
            assert isinstance(cc, ConnectedComponent)

            if not (cc.strID() in highlight_vertices):
                self.__image_render_cc(result, cc, color_cc_normal)

        # render highlighted edges ...
        for p1, p2, color, width in lines_highlight:
            cv2.line(result, p1, p2, color, thickness=width)

        # render the highlighted CCs ...
        for cc in self.vertices:
            assert isinstance(cc, ConnectedComponent)

            if cc.strID() in highlight_vertices:
                if not color_multicolor_highlight:
                    self.__image_render_cc(result, cc, color_cc_highlight)
                else:
                    idx = highlight_vertices.index(cc.strID())
                    current_color = multi_color_highlight[idx % len(multi_color_highlight)]

                    self.__image_render_cc(result, cc, current_color)

        return result

    def buildUnitVector3DCache(self, force=False):
        if self.cache_3D_vectors is not None and not force:
            # cache is already built and no force rebuilt is asked ...
            return

        self.cache_3D_vectors = {idx: {} for idx in range(len(self.vertices))}
        connected = self.get_connections()
        for cc_1_idx, cc_2_idx in connected:
            cc_1_str = self.vertices[cc_1_idx].strID()
            cc_2_str = self.vertices[cc_2_idx].strID()
            dx, dy, dz = self.pairUnitVector3D(cc_1_str, cc_2_str)

            self.cache_3D_vectors[cc_1_idx][cc_2_idx] = (dx, dy, dz)
            self.cache_3D_vectors[cc_2_idx][cc_1_idx] = (-dx, -dy, dz)

    def pairUnitVector2D(self, str_id_1, str_id_2):
        node_1 = self.vertices[self.vertices_by_id[str_id_1]]
        node_2 = self.vertices[self.vertices_by_id[str_id_2]]

        n1_x, n1_y = node_1.getCenter()
        n2_x, n2_y = node_2.getCenter()

        dx = n2_x - n1_x
        dy = n2_y - n1_y

        norm = math.sqrt(dx * dx + dy * dy)

        if norm > 0.0:
            dx /= norm
            dy /= norm
        else:
            dx = dy = 0.0

        return dx, dy

    def pairUnitVector3D(self, str_id_1, str_id_2):
        v_1 = self.vertices[self.vertices_by_id[str_id_1]]
        v_2 = self.vertices[self.vertices_by_id[str_id_2]]

        x1, y1 = v_1.getCenter()
        x2, y2 = v_2.getCenter()

        dx = x2 - x1
        dy = y2 - y1

        norm = math.sqrt(dx * dx + dy * dy)

        # check if too close (possible containment)
        r_1 = min(v_1.getWidth(), v_1.getHeight()) / 2.0
        r_2 = min(v_2.getWidth(), v_2.getHeight()) / 2.0
        r_p = max(r_1, r_2)
        if norm < r_p:
            # add an spherical factor to simulate 3D ...
            dz = math.sqrt(r_p * r_p - norm * norm)
            # use the simulated sphere radius as the norm ...
            norm = r_p
        else:
            # assume flat
            dz = 0.0

        dx /= norm
        dy /= norm
        dz /= norm

        return dx, dy, dz

    def pairSizeRatio(self, str_id_1, str_id_2):
        v_1 = self.vertices[self.vertices_by_id[str_id_1]]
        v_2 = self.vertices[self.vertices_by_id[str_id_2]]

        diag_1 = v_1.getBoxDiagonal()
        diag_2 = v_2.getBoxDiagonal()

        return diag_1 / diag_2

    @staticmethod
    def spatially_align_pairs(graph_1, pair_1_id, graph_2, pair_2_id):
        g1_node_1_id, g1_node_2_id = pair_1_id.split("|")
        g2_node_1_id, g2_node_2_id = pair_2_id.split("|")

        g1_dx, g1_dy = graph_1.pairUnitVector2D(g1_node_1_id, g1_node_2_id)
        g2_dx, g2_dy = graph_2.pairUnitVector2D(g2_node_1_id, g2_node_2_id)

        cos_sim = g1_dx * g2_dx + g1_dy * g2_dy

        g1_node_1_idx = graph_1.vertices_by_id[g1_node_1_id]
        g1_node_2_idx = graph_1.vertices_by_id[g1_node_2_id]

        g2_node_1_idx = graph_2.vertices_by_id[g2_node_1_id]
        g2_node_2_idx = graph_2.vertices_by_id[g2_node_2_id]

        if cos_sim >= 0.0:
            # first to second is same direction in both pairs ....
            return (g1_node_1_idx, g2_node_1_idx), (g1_node_2_idx, g2_node_2_idx)
        else:
            # opposite direction ..
            return (g1_node_1_idx, g2_node_2_idx), (g1_node_2_idx, g2_node_1_idx)
