
import numpy as np
import itertools
import numpy.lib.recfunctions
import matplotlib.mlab

from TangentV.indexing.index_parameters import IndexParameters

class GraphIndexEntry:
    class EntryInfo:
        def __init__(self, pair_id, conf_1, conf_2, dx, dy, dz, size_ratio, flipped_pair):
            self.pair_id = pair_id

            self.conf_1 = conf_1
            self.conf_2 = conf_2

            self.delta_x = dx
            self.delta_y = dy
            self.delta_z = dz
            self.size_ratio = size_ratio
            self.flipped_pair = flipped_pair


        def get_score_array(self):
            pair_sign = (-1.0 if self.flipped_pair else 1.0)
            return [self.delta_x, self.delta_y, self.delta_z, self.conf_1, self.conf_2, pair_sign, self.size_ratio]

    def __init__(self, label_1, label_2):
        self.label_1 = label_1
        self.label_2 = label_2

        self.instances_per_image = {}

        self.optimized_scores = None
        self.optimized_meta_data = None

    def optimize_for_search(self):
        if self.optimized_scores is not None:
            # already optimized
            return

        # pre-built structure that optimizes queries ...
        score_array = []
        meta_data_array = []
        for img_id in self.instances_per_image:
            for entry in self.instances_per_image[img_id]:
                score_array.append(entry.get_score_array())
                # meta_data_array.append((str(img_id) + "|" + entry.pair_id, img_id, entry.pair_id))
                meta_data_array.append((entry.pair_id, img_id))

        self.optimized_scores = np.array(score_array, dtype=np.float16)
        # ... re normalize 3D vectors after converting from float64 to float16 ... to reduce numerical error later
        norms = np.linalg.norm(self.optimized_scores[:, 0:3], axis=1)
        self.optimized_scores[:, 0] /= norms
        self.optimized_scores[:, 1] /= norms
        self.optimized_scores[:, 2] /= norms

        self.optimized_meta_data = np.array(meta_data_array, dtype=np.int32)
        #print((type(self.optimized_scores), self.optimized_scores.dtype, self.optimized_scores.size, self.optimized_scores.nbytes))
        #print((type(self.optimized_meta_data), self.optimized_meta_data.dtype, self.optimized_meta_data.size, self.optimized_meta_data.nbytes))
        # No more data is expected to come!
        self.instances_per_image = None


    def getID(self):
        return GraphIndexEntry.GenerateID(self.label_1, self.label_2)

    @staticmethod
    def GenerateID(label_1, label_2):
        return label_1 + "|" + label_2

    def add_instance(self, img_id, raw_pair):
        if not img_id in self.instances_per_image:
            self.instances_per_image[img_id] = []

        # create posting list entry ....
        info = GraphIndexEntry.EntryInfo(raw_pair.pair_id, raw_pair.conf_1, raw_pair.conf_2,
                                         raw_pair.delta_x, raw_pair.delta_y, raw_pair.delta_z,
                                         raw_pair.size_ratio, raw_pair.inverted)

        self.instances_per_image[img_id].append(info)

    def total_instances(self):
        total = 0
        for img_id in self.instances_per_image:
            total += len(self.instances_per_image[img_id])

        return total

    def match(self, raw_pair):
        # matching should return a list  of the form ...
        # (img_id, pair_id, score)
        # this could be merged and added externally ...
        cos_sim = self.optimized_scores[:, 0:3] * np.array([[raw_pair.delta_x, raw_pair.delta_y, raw_pair.delta_z]])
        cos_sim = cos_sim.sum(axis=1)

        valid_matches = cos_sim > IndexParameters.IndexEntry_MinCosineSimilarity
        """
        if np.count_nonzero(cos_sim > 1.0):
            positions = np.nonzero(cos_sim > 1.0)
            print(positions)
            print(self.optimized_scores[positions, 0:3])
            print((raw_pair.delta_x, raw_pair.delta_y, raw_pair.delta_z))
            print(cos_sim[positions])
            print(np.linalg.norm( np.array([raw_pair.delta_x, raw_pair.delta_y, raw_pair.delta_z])) )
            print(np.linalg.norm(self.optimized_scores[positions, 0:3]))
            x = 0 / 0
        """
        # check size ratios
        if IndexParameters.IndexEntry_MinSizeRatio > 0.0:
            size_ratios = self.optimized_scores[:, 6]
            pair_ratio = np.array([raw_pair.size_ratio])

            # compute the size-ratio score....
            ratio_scores = np.minimum(size_ratios, pair_ratio) / np.maximum(size_ratios, pair_ratio)

            valid_ratios = ratio_scores > IndexParameters.IndexEntry_MinSizeRatio

            # apply the size-based filter ...
            valid_matches = np.logical_and(valid_matches, valid_ratios)


        # get cosine similarity scores for matching samples
        matching_cos_sim = cos_sim[valid_matches]

        if IndexParameters.IndexEntry_ScaleCosineSimilarity:
            scale_factor = 1 / (1.0 - IndexParameters.IndexEntry_MinCosineSimilarity)
            matching_cos_sim -= IndexParameters.IndexEntry_MinCosineSimilarity
            matching_cos_sim *= scale_factor

            """
            for idx in range(matching_cos_sim.shape[0]):
                if matching_cos_sim[idx] > 1.0:
                    print(IndexParameters.IndexEntry_MinCosineSimilarity)
                    print(scale_factor)
                    print(matching_cos_sim[idx])
                    print(idx)
                    print(cos_sim[valid_matches])
                    x = 0 / 0
            """

        matching_opt_scores = self.optimized_scores[valid_matches]

        if IndexParameters.SimilarityMetric == IndexParameters.Metric_ProductBased:
            # (P(l^1_i| n_i) * P(l^1_q| n_q)) * ( P(l^2_i| n_i) * P(l^2_q| n_q))

            # compute label similarity scores for matching samples
            matching_label_sim = matching_opt_scores[:, 3:5] * np.array([[raw_pair.conf_1, raw_pair.conf_2]])

            # use the product of all confidences ...
            matching_label_sim = matching_label_sim.prod(axis=1)

            # compute final scores for matching samples
            matching_scores = matching_cos_sim * matching_label_sim
        elif IndexParameters.SimilarityMetric == IndexParameters.Metric_MinBased:
            # Min( P(l^1_i| n_i), P(l^1_q| n_q)) * Min( P(l^2_i| n_i), P(l^2_q| n_q))

            # compute label similarity scores for matching samples ... using minimum
            matching_label_sim = np.minimum(matching_opt_scores[:, 3:5], np.array([[raw_pair.conf_1, raw_pair.conf_2]]))

            # use the product of the two minimum confidences for each pair of labels
            matching_label_sim = matching_label_sim.prod(axis=1)

            # compute final scores for matching samples
            matching_scores = matching_cos_sim * matching_label_sim
        elif (IndexParameters.SimilarityMetric == IndexParameters.Metric_MaximumLineOfSightSimilarity or
              IndexParameters.SimilarityMetric == IndexParameters.Metric_MinBasedMLS):
            # only use the cosine similarity metric
            matching_scores = matching_cos_sim
        else:
            raise Exception("Invalid Similarity Metric selected")

        # compute direction agreement ...
        pair_sign = (-1.0 if raw_pair.inverted else 1.0)
        matching_direction = matching_opt_scores[:, 5] * pair_sign

        # extract metadata for matching samples
        matching_meta_data = self.optimized_meta_data[valid_matches]

        # print(self.label_1 + "\t" + self.label_2 + "\t" + str(matching_scores.shape[0]))

        return matching_scores, matching_meta_data, matching_direction

    @staticmethod
    def merge_matches_by_pair(query_pair_scored_matches, debug_img_id=None):
        final_by_pair = {}

        # Merge (aggregate) by unique id of candidate pairs (kf_id + pair_id)
        # TODO: sorting posting lists by unique_ids and using iterator trees should boost performance for this op.
        for array_scores, array_meta_data, array_direction in query_pair_scored_matches:
            for score, (unique_id, img_id), direction in zip(array_scores, array_meta_data, array_direction):
                # move from str to original data type: int
                img_id = int(img_id)
                # print(str(img_id) + "\t" + pair_id + "\t" + str(score) + "\t" + str(direction))
                if not unique_id in final_by_pair:
                    final_by_pair[unique_id] = [img_id, unique_id, score, direction]
                else:
                    pair = final_by_pair[unique_id]
                    pair[3] += direction

                    if ((IndexParameters.SimilarityMetric == IndexParameters.Metric_ProductBased) or
                        (IndexParameters.SimilarityMetric == IndexParameters.Metric_MinBased)):
                        # simply add the scores ....
                        pair[2] += score
                    elif ((IndexParameters.SimilarityMetric == IndexParameters.Metric_MaximumLineOfSightSimilarity) or
                          (IndexParameters.SimilarityMetric == IndexParameters.Metric_MinBasedMLS)):
                        # Do nothing, score should contain only one value with cosine similarity for alignment
                        # which should be identical for all combinations.
                        pass
                    else:
                        raise Exception("Unknown similarity metric")

                """
                # commented for performance, un-comment for debugging
                if img_id == debug_img_id:
                    print(str(score) + "\t" + pair_id + "\t" + str(direction))
                """

        final_by_pair = final_by_pair.values()

        return final_by_pair
