
from TangentV.data.spatial_cc_graph import SpatialCCGraph
from TangentV.indexing.index_parameters import IndexParameters

class GraphMatch:
    DebugCounter = 0
    Err_Constant = 0.01

    def __init__(self, query_graph, candidate_graph):
        self.query_graph = query_graph
        self.candidate_graph = candidate_graph
        self.query_nodes = set()
        self.candidate_nodes = set()
        self.node_alignments = set()
        self.pair_alignments = []
        self.scores = []
        self.cache_weighted_recall = {}
        self.total_score = 0.0
        # TODO: keep Copy method up to date if anything gets added here ...

    @staticmethod
    def Copy(match):
        assert isinstance(match, GraphMatch)

        copy_match = GraphMatch(match.query_graph, match.candidate_graph)
        copy_match.query_nodes = set(match.query_nodes)
        copy_match.candidate_nodes = set(match.candidate_nodes)
        copy_match.node_alignments = set(match.node_alignments)
        copy_match.pair_alignments = list(match.pair_alignments)
        copy_match.scores = list(match.scores)
        copy_match.cache_weighted_recall = dict(match.cache_weighted_recall)
        copy_match.total_score = match.total_score

        return copy_match

    def candidate_region(self):
        all_mins_x, all_maxs_x, all_mins_y, all_maxs_y = [], [], [], []
        for cc_idx in self.candidate_nodes:
            all_mins_x.append(self.candidate_graph.vertices[cc_idx].min_x)
            all_maxs_x.append(self.candidate_graph.vertices[cc_idx].max_x)
            all_mins_y.append(self.candidate_graph.vertices[cc_idx].min_y)
            all_maxs_y.append(self.candidate_graph.vertices[cc_idx].max_y)

        final_min_x = min(all_mins_x)
        final_max_x = max(all_maxs_x)
        final_min_y = min(all_mins_y)
        final_max_y = max(all_maxs_y)

        return final_min_x, final_max_x, final_min_y, final_max_y

    def strID(self):
        return "-".join(["({0:d},{1:d})".format(u, v) for u, v in sorted(self.node_alignments)])

    def __str__(self):
        str_q_nodes = ", ".join([str(u) for u in sorted(self.query_nodes)])
        str_c_nodes = ", ".join([str(v) for v in sorted(self.candidate_nodes)])
        str_alignments = "-".join(["({0:d},{1:d})".format(u, v) for u, v in sorted(self.node_alignments)])
        str_scores = ", ".join([str(score) for score in self.scores])
        str_pairs = ",\n".join(["\t\t" + q_id + "\t" + c_id for q_id, c_id in self.pair_alignments])

        template = "<Match: ID='{0:s}',\n\tScore={1:.4f},\n\tQ=[{2:s}],\n\tC=[{3:s}],\n\tNode-Aligned=[{4:s}]," \
                   "\n\tPairs-Scores=[{5:s}],\n\tPairs-Aligned=[\n{6:s}\n\t]>"
        return template.format(self.strID(), self.total_score, str_q_nodes, str_c_nodes, str_alignments,
                               str_scores, str_pairs)

    def __repr__(self):
        return str(self)

    def print_rec_confidences(self):
        print("Match: ID:{0:s} Aligned Vertex Label Confidences".format(self.strID()))
        for q_node_idx, c_node_idx in self.node_alignments:
            q_str_id = self.query_graph.vertices[q_node_idx].strID()
            c_str_id = self.candidate_graph.vertices[c_node_idx].strID()

            # vertices_labels, vertices_confidences
            print("\tQ: {0:s} - C: {1:s}".format(q_str_id, c_str_id))

            # first, add classes from query ...
            unique_classes = {}
            for idx, label in enumerate(self.query_graph.vertices_labels[q_node_idx]):
                confidence = self.query_graph.vertices_confidences[q_node_idx][idx]
                unique_classes[label] = {"query": confidence, "candidate": None, 'relevance': confidence}
            # then, from the candidate ...
            for idx, label in enumerate(self.candidate_graph.vertices_labels[c_node_idx]):
                confidence = self.candidate_graph.vertices_confidences[c_node_idx][idx]
                if label in unique_classes:
                    unique_classes[label]["candidate"] = confidence
                    unique_classes[label]["relevance"] += confidence
                else:
                    unique_classes[label] = {"query": None, "candidate": confidence, 'relevance': confidence}

            label_order = sorted([(unique_classes[l]['relevance'], l) for l in unique_classes], reverse=True)

            # print them side-ways ...
            print("\t\tQ %\tC %\tLabel")
            for relevance, label in label_order:
                if unique_classes[label]["query"] is None:
                    query_val = "---"
                else:
                    query_val = "{0:.2f}".format(unique_classes[label]["query"] * 100.0)

                if unique_classes[label]["candidate"] is None:
                    candidate_val = "---"
                else:
                    candidate_val = "{0:.2f}".format(unique_classes[label]["candidate"] * 100.0)

                print("\t\t{0:s}\t{1:s}\t{2:s}".format(query_val, candidate_val, label))



    def check_compatibility(self, other_match):
        # assert isinstance(other_match, GraphMatch)

        local_align_set = self.node_alignments
        other_align_set = other_match.node_alignments

        common_align_set = local_align_set.intersection(other_align_set)

        common_nodes_query = self.query_nodes.intersection(other_match.query_nodes)
        common_nodes_candidate = self.candidate_nodes.intersection(other_match.candidate_nodes)

        if (len(common_align_set) == 0 or len(common_nodes_query) > len(common_align_set) or
            len(common_nodes_candidate) > len(common_align_set)):
            # Not compatible if there are no aligned nodes in common ...
            return False, 0.0
        else:
            edge_cost = self.total_score + other_match.total_score
            return True, edge_cost

    def check_partial_compatibility(self, other_match):
        # check if other has alignments that could be added to this match ..
        diff_query = other_match.query_nodes - self.query_nodes

        if len(diff_query) == 0:
            return []
        else:
            # check for alignments that are missing in current match ...
            missing_alignments = []
            for qn_idx, cn_idx in other_match.node_alignments:
                if (not qn_idx in self.query_nodes) and (not cn_idx in self.candidate_nodes):
                    missing_alignments.append((qn_idx, cn_idx))

            # check for identical alignments ...
            common_alignments = list(self.node_alignments.intersection(other_match.node_alignments))

            return missing_alignments + common_alignments

    def check_disjoint_alignments(self, other_match):
        # assert isinstance(other_match, GraphMatch)

        common_nodes_query = self.query_nodes.intersection(other_match.query_nodes)
        common_nodes_candidate = self.candidate_nodes.intersection(other_match.candidate_nodes)

        if len(common_nodes_query) > 0 or len(common_nodes_candidate) > 0:
            # Not disjoint if there are common query or candidate nodes ...
            return False, 0.0
        else:
            edge_cost = self.total_score + other_match.total_score
            return True, edge_cost

    def getImage(self, highlight_nodes=True, use_multicolor=True, show_edges=True):
        assert isinstance(self.query_graph, SpatialCCGraph)

        hl_q_edges, hl_c_edges = [], []
        for query_pair_id, candidate_pair_id in self.pair_alignments:
            q_id_1, q_id_2 = query_pair_id.split("|")
            c_id_1, c_id_2 = candidate_pair_id.split("|")

            hl_q_edges.append((q_id_1, q_id_2))
            hl_c_edges.append((c_id_1, c_id_2))

        hl_q_vertices = []
        hl_c_vertices = []
        if highlight_nodes:
            for q_node_idx, kf_node_idx in self.node_alignments:
                hl_q_vertices.append(self.query_graph.vertices[q_node_idx].strID())
                hl_c_vertices.append(self.candidate_graph.vertices[kf_node_idx].strID())

        style = {"use_multicolor": use_multicolor, "draw_edges": show_edges}
        # render query image ...
        query_match = self.query_graph.render(highlight_vertices=hl_q_vertices, highlight_connections=hl_q_edges,
                                              style=style)

        # render candidate image ...
        candidate_match = self.candidate_graph.render(highlight_vertices=hl_c_vertices,
                                                      highlight_connections=hl_c_edges, style=style)

        return query_match, candidate_match

    def AddPath(self, path, path_scores):
        # assumes all edges are new ...
        if len(path) < 2:
            # no edges specified by path ..
            return

        if len(path_scores) != len(path) - 1:
            raise Exception("Inconsistent number of scores specified for path")

        # first, add nodes ...
        # note these are set operations, elements won't be added unless they are new
        for next_q_idx, next_c_idx in path:
            # to sets of individual nodes ...
            self.query_nodes.add(next_q_idx)
            self.candidate_nodes.add(next_c_idx)
            # to the set of aligned nodes
            self.node_alignments.add((next_q_idx, next_c_idx))

        # now, add edges ...
        last_q_idx, last_c_idx = path[0]
        last_q_str_id = self.query_graph.vertices[last_q_idx].strID()
        last_c_str_id = self.candidate_graph.vertices[last_c_idx].strID()
        for idx, (next_q_idx, next_c_idx) in enumerate(path[1:]):
            next_q_str_id = self.query_graph.vertices[next_q_idx].strID()
            next_c_str_id = self.candidate_graph.vertices[next_c_idx].strID()

            # define pair ids ....
            if last_q_idx < next_q_idx:
                query_pair_id = last_q_str_id + "|" + next_q_str_id
            else:
                query_pair_id = next_q_str_id + "|" + last_q_str_id

            if last_c_idx < next_c_idx:
                candidate_pair_id = last_c_str_id + "|" + next_c_str_id
            else:
                candidate_pair_id = next_c_str_id + "|" + last_c_str_id

            self.pair_alignments.append((query_pair_id, candidate_pair_id))
            self.scores.append(path_scores[idx])

            # if pair-based, simply accumulate score
            if ((IndexParameters.SimilarityMetric == IndexParameters.Metric_ProductBased) or
                (IndexParameters.SimilarityMetric == IndexParameters.Metric_MinBased)):
                self.total_score += path_scores[idx]

            last_q_idx, last_c_idx = next_q_idx, next_c_idx
            last_q_str_id, last_c_str_id = next_q_str_id, next_c_str_id

        if IndexParameters.SimilarityMetric == IndexParameters.Metric_MaximumLineOfSightSimilarity:
            # Line of Sight metric requires to be re-computed ...
            self.total_score = self.getMaximumLOSSimilarity()
        elif IndexParameters.SimilarityMetric == IndexParameters.Metric_MinBasedMLS:
            # Line of Sight metric requires to be re-computed ...
            self.total_score = self.getMinBasedMaximumLOSSimilarity()

    def getAlignmentMatchedLabels(self, q_node_idx, img_node_idx):
        q_node_classes = self.query_graph.vertices_labels[q_node_idx]
        q_node_confidences = self.query_graph.vertices_confidences[q_node_idx]
        c_node_classes = self.candidate_graph.vertices_labels[img_node_idx]
        c_node_confidences = self.candidate_graph.vertices_confidences[img_node_idx]

        if q_node_classes[0] == SpatialCCGraph.WildcardLabel:
            # wildcard node match (100% confidence for any label combination)
            return {SpatialCCGraph.WildcardLabel: (1.0, 1.0)}

        c_inverted_conf = {}
        for idx, c_class_label in enumerate(c_node_classes):
            c_conf = c_node_confidences[idx]
            c_inverted_conf[c_class_label] = c_conf

        matching_labels = {}
        for idx, q_class_label in enumerate(q_node_classes):
            if q_class_label in c_inverted_conf:
                q_conf = q_node_confidences[idx]
                matching_labels[q_class_label] = (q_conf, c_inverted_conf[q_class_label])

        return matching_labels

    def getProdBasedAlignmentNodeRecall(self, q_node_idx, img_node_idx):
        matched_labels = self.getAlignmentMatchedLabels(q_node_idx, img_node_idx)

        total_weight = 0.0
        for label in matched_labels:
            q_conf, c_conf = matched_labels[label]
            total_weight += q_conf * c_conf

        return total_weight

    def getMinBasedAlignmentNodeRecall(self, q_node_idx, img_node_idx):
        matched_labels = self.getAlignmentMatchedLabels(q_node_idx, img_node_idx)

        total_weight = 0.0
        for label in matched_labels:
            q_conf, c_conf = matched_labels[label]
            total_weight += min(q_conf, c_conf)

        return total_weight

    def getProdBasedWeightedNodeRecall(self):
        weighted_recall = 0.0
        for q_node_idx, img_node_idx in self.node_alignments:
            alignment_key = str(q_node_idx) + "|" + str(img_node_idx)
            if not alignment_key in self.cache_weighted_recall:
                self.cache_weighted_recall[alignment_key] = self.getProdBasedAlignmentNodeRecall(q_node_idx, img_node_idx)

            weighted_recall += self.cache_weighted_recall[alignment_key]

        weighted_recall /= len(self.query_graph.vertices)

        return weighted_recall

    def getMinBasedWeightedNodeRecall(self):
        weighted_recall = 0.0
        for q_node_idx, img_node_idx in self.node_alignments:
            alignment_key = str(q_node_idx) + "|" + str(img_node_idx)
            if not alignment_key in self.cache_weighted_recall:
                self.cache_weighted_recall[alignment_key] = self.getMinBasedAlignmentNodeRecall(q_node_idx, img_node_idx)

            weighted_recall += self.cache_weighted_recall[alignment_key]

        # weighted recall is unlikely to be 100% unless sym rec is perfect!
        weighted_recall /= len(self.query_graph.vertices)

        return weighted_recall

    def getMaximumLOSSimilarity(self):
        # assume that "scores" holds the cosine similarity for all pairs ... then ...
        edge_weighted_recall = sum(self.scores) / self.query_graph.total_connections()
        node_weighted_recall = self.getProdBasedWeightedNodeRecall()

        if 1.0 < edge_weighted_recall < 1.0 + GraphMatch.Err_Constant:
            # correct small float precision numeric errors ...
            edge_weighted_recall = 1.0

        if edge_weighted_recall + node_weighted_recall > 0.0:
            MLS = (2.0 * edge_weighted_recall * node_weighted_recall) / (edge_weighted_recall + node_weighted_recall)
        else:
            MLS = 0.0

        if not (-GraphMatch.Err_Constant <= MLS <= 1.0 + GraphMatch.Err_Constant):
            print(str(self))
            print("Weighted Edge Recall: " + str(edge_weighted_recall))
            print("Weighted Node Recall: " + str(node_weighted_recall))
            print(self.cache_weighted_recall)
            raise Exception("Invalid ranges for weighted node/edge recall")

        # GraphMatch.DebugCounter += 1
        # print((GraphMatch.DebugCounter, MLS), flush=True)

        return MLS

    def getMinBasedMaximumLOSSimilarity(self):
        # assume that "scores" holds the cosine similarity for all pairs ... then ...
        edge_weighted_recall = sum(self.scores) / self.query_graph.total_connections()
        node_weighted_recall = self.getMinBasedWeightedNodeRecall()

        if 1.0 < edge_weighted_recall < 1.0 + GraphMatch.Err_Constant:
            # correct small float precision numeric errors ...
            edge_weighted_recall = 1.0

        if edge_weighted_recall + node_weighted_recall > 0.0:
            MLS = (2.0 * edge_weighted_recall * node_weighted_recall) / (edge_weighted_recall + node_weighted_recall)
        else:
            MLS = 0.0

        if not (-GraphMatch.Err_Constant <= MLS <= 1.0 + GraphMatch.Err_Constant):
            print(str(self))
            print("Weighted Edge Recall: " + str(edge_weighted_recall))
            print("Weighted Node Recall: " + str(node_weighted_recall))
            print(self.cache_weighted_recall)
            print(MLS)
            raise Exception("Invalid ranges for weighted node/edge recall")

        # GraphMatch.DebugCounter += 1
        # print((GraphMatch.DebugCounter, MLS), flush=True)

        return MLS


    def getGlobalPairBasedPrecision(self):
        matched_pairs = len(self.pair_alignments)

        return matched_pairs / self.candidate_graph.total_connections()

    @staticmethod
    def SubMatch(match, alignments):
        assert isinstance(match, GraphMatch)

        if not isinstance(alignments, set):
            alignments = set(alignments)

        valid_match_alignments = match.node_alignments.intersection(alignments)
        if len(valid_match_alignments) == len(match.node_alignments):
            # same length, nothing to cut ...
            return match
        else:
            # filter ...
            graph_match = GraphMatch(match.query_graph, match.candidate_graph)
            graph_match.node_alignments = valid_match_alignments

            for q_node, c_node in valid_match_alignments:
                graph_match.query_nodes.add(q_node)
                graph_match.candidate_nodes.add(c_node)

            for idx, (q_pair_id, c_pair_id) in enumerate(match.pair_alignments):
                q_node_id_1, q_node_id_2 = q_pair_id.split("|")
                c_node_id_1, c_node_id_2 = c_pair_id.split("|")

                if (match.query_graph.vertices_by_id[q_node_id_1] in graph_match.query_nodes and
                    match.query_graph.vertices_by_id[q_node_id_2] in graph_match.query_nodes and
                    match.candidate_graph.vertices_by_id[c_node_id_1] in graph_match.candidate_nodes and
                    match.candidate_graph.vertices_by_id[c_node_id_2] in graph_match.candidate_nodes):
                    # pair is valid ..
                    graph_match.pair_alignments.append((q_pair_id, c_pair_id))
                    graph_match.scores.append(match.scores[idx])

                    if ((IndexParameters.SimilarityMetric == IndexParameters.Metric_ProductBased) or
                        (IndexParameters.SimilarityMetric == IndexParameters.Metric_MinBased)):
                        # Simply accumulate ...
                        graph_match.total_score += match.scores[idx]

            if IndexParameters.SimilarityMetric == IndexParameters.Metric_MaximumLineOfSightSimilarity:
                # re-compute score
                graph_match.total_score = graph_match.getMaximumLOSSimilarity()
            elif IndexParameters.SimilarityMetric == IndexParameters.Metric_MinBasedMLS:
                graph_match.total_score = graph_match.getMinBasedMaximumLOSSimilarity()

            return graph_match

    @staticmethod
    def Merge(match_1, match_2):
        assert isinstance(match_1, GraphMatch) and isinstance(match_2, GraphMatch)

        graph_match = GraphMatch(match_1.query_graph, match_1.candidate_graph)
        graph_match.query_nodes = set(match_1.query_nodes).union(set(match_2.query_nodes))
        graph_match.candidate_nodes = set(match_1.candidate_nodes).union(set(match_2.candidate_nodes))
        graph_match.node_alignments = set(match_1.node_alignments).union(set(match_2.node_alignments))

        # these are not merged in case there are repetitions, which must be filtered ...
        graph_match.pair_alignments = match_1.pair_alignments + match_2.pair_alignments
        graph_match.scores = match_1.scores + match_2.scores

        if len(set(graph_match.pair_alignments)) < len(graph_match.pair_alignments):
            # merge matched pairs and keep their corresponding scores consistent
            # this allows to merge matches with overlap in matching pairs ...
            temporal_unique_pairs = {}
            for idx, (query_pair_id, pair_id) in enumerate(graph_match.pair_alignments):
                temporal_unique_pairs[query_pair_id + pair_id] = ((query_pair_id, pair_id), graph_match.scores[idx])

            temporal_alignments = []
            temporal_scores = []
            for alignment, score in list(temporal_unique_pairs.values()):
                temporal_alignments.append(alignment)
                temporal_scores.append(score)

            graph_match.pair_alignments = temporal_alignments
            graph_match.scores = temporal_scores

        if ((IndexParameters.SimilarityMetric == IndexParameters.Metric_ProductBased) or
            (IndexParameters.SimilarityMetric == IndexParameters.Metric_MinBased)):
            # pair-based
            graph_match.total_score = sum(graph_match.scores)
        elif IndexParameters.SimilarityMetric == IndexParameters.Metric_MaximumLineOfSightSimilarity:
            graph_match.total_score = graph_match.getMaximumLOSSimilarity()
        elif IndexParameters.SimilarityMetric == IndexParameters.Metric_MinBasedMLS:
            graph_match.total_score = graph_match.getMinBasedMaximumLOSSimilarity()
        else:
            raise Exception("Unknown similarity metric")

        return graph_match

    @staticmethod
    def FromPair(query_graph, query_pair_id, candidate_graph, pair_id, score, direction):
        # use the information provided from the index ...
        q_n1_str, q_n2_str = query_pair_id.split("|")
        c_n1_str, c_n2_str = pair_id.split("|")

        if direction >= 0:
            m1 = (query_graph.vertices_by_id[q_n1_str], candidate_graph.vertices_by_id[c_n1_str])
            m2 = (query_graph.vertices_by_id[q_n2_str], candidate_graph.vertices_by_id[c_n2_str])
        else:
            m1 = (query_graph.vertices_by_id[q_n1_str], candidate_graph.vertices_by_id[c_n2_str])
            m2 = (query_graph.vertices_by_id[q_n2_str], candidate_graph.vertices_by_id[c_n1_str])

        (q_node_1_idx, img_node_1_idx), (q_node_2_idx, img_node_2_idx) = m1, m2

        graph_match = GraphMatch(query_graph, candidate_graph)

        # initialize with matching pair ...
        graph_match.query_nodes = {q_node_1_idx, q_node_2_idx}
        graph_match.candidate_nodes = {img_node_1_idx, img_node_2_idx}

        graph_match.node_alignments = {(q_node_1_idx, img_node_1_idx), (q_node_2_idx, img_node_2_idx)}
        graph_match.pair_alignments = [(query_pair_id, pair_id)]

        # these vectors will have different meaning based on similarity metric
        # for ProductBased, it will be the individual score (confidence and cosine sim product)
        # for MLS, it will only keep the cosine similarity for each pair ... the whole score would need to be
        #    recomputed when the graph changes ..
        graph_match.scores = [score]

        if ((IndexParameters.SimilarityMetric == IndexParameters.Metric_ProductBased) or
            (IndexParameters.SimilarityMetric == IndexParameters.Metric_MinBased)):
            # pair-based scoring metrics ...
            graph_match.total_score = score
        elif IndexParameters.SimilarityMetric == IndexParameters.Metric_MaximumLineOfSightSimilarity:
            graph_match.total_score = graph_match.getMaximumLOSSimilarity()
        elif IndexParameters.SimilarityMetric == IndexParameters.Metric_MinBasedMLS:
            graph_match.total_score = graph_match.getMinBasedMaximumLOSSimilarity()
        else:
            raise Exception("Unknown similarity metric")

        return graph_match

    @staticmethod
    def region_overlap(region_1, region_2):
        r1_min_x, r1_max_x, r1_min_y, r1_max_y = region_1
        r2_min_x, r2_max_x, r2_min_y, r2_max_y = region_2

        # first, check if they overlap
        if r1_min_x < r2_max_x and r2_min_x < r1_max_x and r1_min_y < r2_max_y and r2_min_y < r1_max_y:
            common_min_x = max(r1_min_x, r2_min_x)
            common_max_x = min(r1_max_x, r2_max_x)
            common_min_y = max(r1_min_y, r2_min_y)
            common_max_y = min(r1_max_y, r2_max_y)

            r1_area = (r1_max_x - r1_min_x + 1) * (r1_max_y - r1_min_y + 1)
            r2_area = (r2_max_x - r2_min_x + 1) * (r2_max_y - r2_min_y + 1)
            common_area = (common_max_x - common_min_x + 1) * (common_max_y - common_min_y + 1)

            return common_area, common_area / r1_area, common_area / r2_area
        else:
            return 0.0, 0.0, 0.0

    @staticmethod
    def FindPathConnectingMatches(query_graph, candidate_graph, first_alignment, second_alignment, min_cos_sim, max_path_len):
        # built sets and check for answer already in start set ...
        pre_aligned_query = set()
        pre_aligned_candidate = set()
        for q_node, c_node in first_alignment:
            # answer found
            if (q_node, c_node) in second_alignment:
                return [(q_node, c_node)]

            pre_aligned_query.add(q_node)
            pre_aligned_candidate.add(c_node)

        # initialize search ...
        current_paths = []
        alignments_costs = [{} for v in query_graph.vertices]
        for q_node, c_node in first_alignment:
            current_paths.append((0.0, [(q_node, c_node)], pre_aligned_query, pre_aligned_candidate))
            alignments_costs[q_node][c_node] = 0.0

        # do search ....
        while len(current_paths) > 0:
            cost, path, path_aligned_query, path_aligned_candidate = current_paths[0]
            del current_paths[0]

            last_alignment = path[-1]

            last_q_node, last_c_node = last_alignment
            if cost > alignments_costs[last_q_node][last_c_node]:
                # this path has been visited before using shorter path (and solution was not found then)
                # skip ...
                continue

            if last_alignment in second_alignment:
                # path found!
                return path
            else:
                if len(path) == max_path_len:
                    # max length of path reached, do not expand ...
                    continue

                # vertices that have not been visited by current path ...
                q_expansions = set(query_graph.connections[last_q_node, :].nonzero()[0]) - path_aligned_query

                if len(q_expansions) == 0:
                    # cannot further expand this path ...
                    continue

                # find connections not yet in use (candidate)
                c_expansions = set(candidate_graph.connections[last_c_node, :].nonzero()[0]) - path_aligned_candidate
                if len(c_expansions) == 0:
                    # cannot expand this path
                    continue

                # free nodes found, check all pairwise alignments ..
                paths_added = False
                for next_q_node in q_expansions:
                    next_q_dx, next_q_dy, next_q_dz = query_graph.cache_3D_vectors[last_q_node][next_q_node]

                    for next_c_node in c_expansions:
                        next_c_dx, next_c_dy, next_c_dz = candidate_graph.cache_3D_vectors[last_c_node][next_c_node]

                        cos_sim = next_q_dx * next_c_dx + next_q_dy * next_c_dy + next_q_dz * next_c_dz

                        if cos_sim > min_cos_sim:
                            # a valid alignment is possible ...
                            last_node = candidate_graph.vertices[last_c_node]
                            next_node = candidate_graph.vertices[next_c_node]
                            next_distance = last_node.getCenterDistance(next_node)

                            next_cost = cost + next_distance

                            # check if alignment not yet reached or if current cost is lower ....
                            next_q_alignments = alignments_costs[next_q_node]
                            if not next_c_node in next_q_alignments or next_cost < next_q_alignments[next_c_node]:
                                #print("Yes!")
                                next_q_alignments[next_c_node] = next_cost

                                next_path = path + [(next_q_node, next_c_node)]

                                # note that here "Union" is used to add the note to a copy of the set ...
                                next_aligned_query = path_aligned_query.union({next_q_node})
                                next_aligned_candidate = path_aligned_candidate.union({next_c_node})

                                current_paths.append((next_cost, next_path, next_aligned_query, next_aligned_candidate))
                                paths_added = True
                            # else:
                            #    print("NO!")

                if paths_added:
                    # sort in ascending cost ...
                    current_paths = sorted(current_paths, key=lambda x:x[0])

        # No path found!!!
        return []