
import os
import sys
import time
from concurrent.futures import ProcessPoolExecutor

import cv2

from AM_CommonTools.util.time_helper import TimeHelper
from TangentV.util.hw_renderer import HandwritingRenderer

from recognizer.configuration.configuration import Configuration
from recognizer.preprocessing.inkml_loader import INKMLLoader
from recognizer.data.feature_dataset import FeatureDataset

def main():
    # validate params
    if len(sys.argv) < 3:
        print("Usage:")
        print("")
        print("python {0:s} conf_file output_dir".format(sys.argv[0]))
        print("")
        print("Where")
        print("\tconf_file:\tFile with symbol recognizer configuration")
        print("\toutput_dir:\tOutput Directory where rendered traces will be stored ")
        return

    config_filename = sys.argv[1]
    output_directory = sys.argv[2]

    # create the renderer ....
    render_resolution = 500
    render_prc_width = 0.1
    render_decay_width = 0.25
    render_decay_length = 0.10
    render_max_scaled_recay = 2.0

    renderer = HandwritingRenderer(render_resolution, render_prc_width, render_decay_width, render_decay_length,
                                   render_max_scaled_recay)


    start_loading = time.time()

    # load configuration
    config = Configuration.from_file(config_filename)

    # load data
    if config.contains("REPLACE_LABELS"):
        label_replacement = INKMLLoader.load_label_replacement(config.get("REPLACE_LABELS"))
    else:
        label_replacement = {}

    if config.contains("ONLY_ACTIVE_CLASSES"):
        active_classes = config.get("ONLY_ACTIVE_CLASSES")
    else:
        active_classes = None

    workers = config.get_int("PARALLEL_INKML_LOADING_WORKERS", 10)

    all_datasets = [(config.data["TRAINING_DATASET_PATH"], "training"),
                    (config.data["TESTING_DATASET_PATH"], "testing")]

    total_time_loading = 0.0
    total_time_rendering = 0.0
    total_time_storing = 0.0

    label_mapping = None
    class_mapping = None
    for data_path, out_sufix in all_datasets:
        print("Loading {0:s} data from {1:s}".format(out_sufix, data_path))
        task_start_time = time.time()
        data_info = INKMLLoader.load_INKML_directory(data_path, True, None, label_replacement,
                                                     False, None, None, True, workers,
                                                     active_classes)

        dataset, valid_files, error_files = data_info
        (raw_data, raw_labels, sources) = dataset
        print("\nTotal Symbols found: " + str(len(raw_data)))

        # map classes ...
        label_data = FeatureDataset.create_update_label_mapping(raw_labels, label_mapping, class_mapping)
        mapped_labels, label_mapping, class_mapping = label_data

        # create output directories
        # ... parent ...
        if not os.path.isdir(output_directory):
            os.mkdir(output_directory)

        # ... train/test sub ...
        current_dir = output_directory + "/" + out_sufix
        if not os.path.isdir(current_dir):
            os.mkdir(current_dir)

        # ... per class directory ...
        for idx in range(len(label_mapping)):
            label_dir = current_dir + "/" + str(idx)
            if not os.path.isdir(label_dir):
                os.mkdir(label_dir)

        # if training, store mapping index ...
        if out_sufix == "training":
            # mapping in human-readable format ....
            mapping_lines = []
            for class_label in class_mapping:
                class_value = class_mapping[class_label]
                mapping_lines.append(str(class_value) + "\t" + class_label + "\n")

            # store class mapping/index ...
            mapping_filename = output_directory + "/class_mapping.txt"
            with open(mapping_filename, "w", encoding="utf8") as out_file:
                out_file.writelines(mapping_lines)

        task_end_time = time.time()
        total_time_loading += task_end_time - task_start_time

        # generate images for each trace symbol ...
        task_start_time = time.time()
        count_per_class = {}
        with ProcessPoolExecutor(max_workers=workers) as executor:
            for idx, image in enumerate(executor.map(renderer.render_hw_symbol, raw_data)):
                print("Saving rendered image ... " + str(idx + 1), end="\r", flush=True)

                class_key = str(mapped_labels[idx, 0])
                if not class_key in count_per_class:
                    count_per_class[class_key] = 0

                output_filename = "{0:s}/{1:s}/{2:d}.png".format(current_dir, class_key, count_per_class[class_key])
                cv2.imwrite(output_filename, image)

                count_per_class[class_key] += 1

        task_end_time = time.time()
        total_time_rendering += task_end_time - task_start_time

        print("\nAll {0:s} images rendered!".format(out_sufix))

        # releasing memory ...
        raw_data = None

    end_time = time.time()

    print("Total Time Loading: " + TimeHelper.secondsToStr(total_time_loading))
    print("Total Time Rendering: " + TimeHelper.secondsToStr(total_time_rendering))
    print("Overall Time: " + TimeHelper.secondsToStr(end_time - start_loading))

if __name__ == "__main__":
    main()
