
import pickle
import sys
import time

from AccessMath.data.meta_data_DB import MetaDataDB
from AccessMath.indexing.lecture_class_mapping import LectureClassMapping
from AccessMath.indexing.tangents_helper import TangentS_Helper

from TangentS.utility.control import Control
from recognizer.configuration.configuration import Configuration
from recognizer.preprocessing.inkml_loader import INKMLLoader
from recognizer.training.trainer import ClassifierTrainer

def main():
    if len(sys.argv) < 2:
        print("Usage:")
        print("")
        print(" python {0:s} database".format(sys.argv[0]))
        print("")
        print("Where")
        print("\tdatabase:\tAccessMath Database XML file")
        return

    db_filename = sys.argv[1]

    # Load database configuration ...
    try:
        am_database = MetaDataDB.from_file(db_filename)
    except:
        print("Invalid AccessMath database")
        return

    rec_config_filename = am_database.indexing.recognizer_config_filename
    tan_config_filename = am_database.indexing.symbolic_mir_config_filename

    similar_shape_filename = am_database.indexing.mapping_sim_shape_filename
    rec_to_latex_filename = am_database.indexing.mapping_rec_to_latex_filename
    notes_to_rec_filename = am_database.indexing.mapping_notes_to_rec_filename
    to_ignore_notes_filename = am_database.indexing.mapping_to_ignore_notes_filename

    try:
        rec_config = Configuration.from_file(rec_config_filename)
    except:
        print("Invalid Symbol Recognizer configuration")
        return

    # tangent-s control
    try:
        tan_control = Control(tan_config_filename) # control file name (after indexing)
    except:
        print("Invalid tangent-s control file ")
        return

    # For Latex-mapping (will be rendered using this string, but remain as the original class)
    latex_corrections = INKMLLoader.load_label_replacement(rec_to_latex_filename)
    am_labels = LectureClassMapping.load_ordinal_labels(notes_to_rec_filename)

    if to_ignore_notes_filename is not None:
        ign_labels = LectureClassMapping.load_ordinal_labels(to_ignore_notes_filename)
    else:
        ign_labels = {}

    print("Loading Raw Training Set")
    training_set = ClassifierTrainer.load_training_set(rec_config)

    # Extract Unique Classes
    unique_labels = sorted(list(set(training_set.labels)))

    print("Initial Unique labels found: " + str(len(unique_labels)))
    print(unique_labels)

    # Load and apply shape-based corrections (reduces Total classes in CROHME)
    cor_labels = INKMLLoader.load_label_replacement(similar_shape_filename)

    for idx in range(len(unique_labels)):
        # check if replace ...
        if unique_labels[idx] in cor_labels:
            # replace ...
            unique_labels[idx] = cor_labels[unique_labels[idx]]

    # re-compute unique ...
    unique_labels = sorted(list(set(unique_labels)))
    print("Unique after mapping by shape: " + str(len(unique_labels)))
    print(unique_labels)

    # find a mapping between lectures and transcript data ...
    print("-> Loading and mapping lecture notes ....")
    lecture_mapping = TangentS_Helper.map_lectures_to_doc_ids(am_database, tan_control)

    print("A total of {0:d} lectures were successfully mapped to their transcripts".format(len(lecture_mapping)))

    # get the math expressions on each lecture ...
    exp_index = TangentS_Helper.load_index_expressions(tan_control)
    total_exp_count = 0
    for idx in exp_index:
        print("-> {0:s} expressions: {1:d}".format(lecture_mapping[idx].title, len(exp_index[idx])))
        total_exp_count += len(exp_index[idx])

    print("A total of {0:d} math expressions where found in {1:d} lectures".format(total_exp_count, len(exp_index)))

    classes_per_lecture = {}
    for idx in exp_index:
        lecture_title = lecture_mapping[idx].title
        print("\nProcessing: " + lecture_title, flush=True)

        task_start = time.time()
        # first, get the set of unique symbols from CROHME that appear on this lecture (according to transcripts)
        lecture_symbols, _ = TangentS_Helper.get_unique_symbols(exp_index[idx], unique_labels, am_labels, ign_labels)

        print(lecture_symbols)
        classes_per_lecture[lecture_title] = lecture_symbols

    # Create Mapping Holder class ...
    full_mapping = LectureClassMapping(lecture_mapping, unique_labels, classes_per_lecture)
    full_mapping.mapping_sim_shape = cor_labels
    full_mapping.mapping_am_to_rec = am_labels
    full_mapping.mapping_am_ignore = ign_labels
    full_mapping.mapping_rec_to_latex = latex_corrections

    # save to file ...
    out_filename = rec_config.data["CLASSIFIER_DIR"] + "/per_lecture_classes.dat"
    with open(out_filename, "wb") as out_file:
        pickle.dump(full_mapping, out_file, pickle.HIGHEST_PROTOCOL)

    print("Full mapping saved to file ... ")

if __name__ == '__main__':
    main()
