
import codecs
import os
import pickle
import sys
import time

import numpy as np

from AM_CommonTools.util.time_helper import TimeHelper

from AccessMath.data.meta_data_DB import MetaDataDB
from AccessMath.indexing.lecture_class_mapping import LectureClassMapping

from TangentV.indexing.recognition_helper import RecognitionHelper

from recognizer.configuration.configuration import Configuration
from recognizer.data.feature_dataset import FeatureDataset
from recognizer.training.trainer import ClassifierTrainer
from recognizer.evaluation.eval_ops import EvalOps


def main():
    if len(sys.argv) < 2:
        print("Usage:")
        print("")
        print(" python {0:s} database [debug_load]".format(sys.argv[0]))
        print("")
        print("Where")
        print("\tdatabase:\tAccessMath Database XML file")
        print("\tdebug_load:\tDebug mode, load existing classifiers for evaluation")
        return

    db_filename = sys.argv[1]

    # check for debug mode ...
    if len(sys.argv) >= 3:
        debug_load_classifier = int(sys.argv[2]) > 0
    else:
        debug_load_classifier = False

    # Load database configuration ...
    try:
        am_database = MetaDataDB.from_file(db_filename)
    except:
        print("Invalid AccessMath database")
        return

    # Junk modes
    # 0 - Do not train with junk, find confidence threshold
    # 1 - Train with junk, use junk class explicitly
    junk_mode = am_database.indexing.use_explict_junk
    min_threshold_recall = am_database.indexing.implicit_junk_threshold

    rec_conf_filename = am_database.indexing.recognizer_config_filename

    named_classifiers_filename = am_database.indexing.named_hw_classifiers_filename

    start_time = time.time()

    # loading
    # ... symbol recognizer configuration ...
    try:
        rec_config = Configuration.from_file(rec_conf_filename)
    except:
        print("Invalid Symbol Recognizer configuration")
        return

    # ... existing Full Mapping
    mapping_filename = rec_config.data["CLASSIFIER_DIR"] + "/per_lecture_classes.dat"
    print("Loading: " + mapping_filename)

    if os.path.isfile(mapping_filename):
        with open(mapping_filename, "rb") as in_file:
            full_mapping = pickle.load(in_file)
            assert isinstance(full_mapping, LectureClassMapping)
    else:
        print("Lecture-Mappings not found, must be created first")
        return

    # load recognizer data ...
    print("Loading Raw Training Set")
    training_set = ClassifierTrainer.load_training_set(rec_config)
    assert isinstance(training_set, FeatureDataset)
    print("Training data: " + str(training_set.data.shape))

    print("Loading Raw Testing Set")
    testing_set = ClassifierTrainer.load_testing_set(rec_config)
    assert isinstance(testing_set, FeatureDataset)
    print("Testing data: " + str(testing_set.data.shape))

    # correct labels
    RecognitionHelper.correct_training_data_classes(training_set, full_mapping.mapping_sim_shape, True)
    # ... on testing ...
    RecognitionHelper.correct_training_data_classes(testing_set, full_mapping.mapping_sim_shape, False)

    end_loading_time = time.time()

    named_classifiers = []

    base_classifier_name = rec_config.data["CLASSIFIER_DIR"] + "/" + rec_config.data["CLASSIFIER_NAME"]
    print("Training general classifier")

    # .... make a copy of the original labels ....
    original_train_labels = training_set.labels
    original_test_labels = testing_set.labels
    # ... Train/Test a classifier using the provided data and classifier configuration
    # ... note that this operation overrides the source labels using re-mapped labels ...
    classifier = RecognitionHelper.get_reference_training_metrics(training_set, testing_set, rec_config)
    # ... now, restore original labels on data ...
    training_set.labels = original_train_labels
    testing_set.labels = original_test_labels

    out_filename = base_classifier_name + ".dat"
    named_classifiers.append(("default", rec_conf_filename, out_filename))
    out_file = open(out_filename, 'wb')
    pickle.dump(classifier, out_file, pickle.HIGHEST_PROTOCOL)
    out_file.close()

    total_lecture_overhead = 0.0
    total_lecture_training = 0.0
    total_lecture_testing = 0.0
    total_lecture_saving = 0.0

    all_train_acc, all_train_avg, all_train_std = [], [], []
    all_test_acc, all_test_avg, all_test_std = [], [], []
    all_junk_train, all_junk_test = [], []

    # find unique Symbols per lecture ...
    for idx in full_mapping.lecture_mapping:
        lecture_title = full_mapping.lecture_mapping[idx].title
        print("\nProcessing: " + lecture_title, flush=True)

        classifier_filename = base_classifier_name + "_" + lecture_title + ".dat"

        task_start = time.time()
        # first, get the set of unique symbols from CROHME that appear on this lecture (according to transcripts)
        lecture_symbols = full_mapping.classes_per_lecture[lecture_title]

        # split datasets based on the symbols found in the lecture ...
        train_ds = RecognitionHelper.split_training_set_by_classes(training_set, lecture_symbols)
        test_ds = RecognitionHelper.split_training_set_by_classes(testing_set, lecture_symbols)

        lecture_train_dataset, extra_train_dataset = train_ds
        lecture_test_dataset, extra_test_dataset = test_ds

        # print some stats....
        n_training_samples, n_training_features = lecture_train_dataset.data.shape
        n_extra_samples = extra_train_dataset.data.shape[0]
        prc_training = n_training_samples / training_set.data.shape[0]
        prc_extra = n_extra_samples / training_set.data.shape[0]

        n_testing_samples = lecture_test_dataset.data.shape[0]
        n_testing_extra_samples = extra_test_dataset.data.shape[0]
        prc_testing = n_testing_samples / testing_set.data.shape[0]
        prc_testing_extra = n_testing_extra_samples / testing_set.data.shape[0]

        print("-> Total features: {0:d}".format(n_training_features))

        print("-> Training Symbols")
        print("---> from Lecture Classes: {0:d} ({1:.2f}%)".format(n_training_samples, prc_training * 100.0))
        print("---> from Other Classes: {0:d} ({1:.2f}%)".format(n_extra_samples, prc_extra * 100.0))
        print("-> Testing Symbols")
        print("---> from Lecture Classes: {0:d} ({1:.2f}%)".format(n_testing_samples, prc_testing * 100.0))
        print("---> from Other Classes: {0:d} ({1:.2f}%)".format(n_testing_extra_samples, prc_testing_extra * 100.0))

        # update label mappings on subset datasets
        if junk_mode == 1:
            # add extra data as JUNK class to training dataset
            lecture_train_dataset.data = np.vstack((lecture_train_dataset.data, extra_train_dataset.data))
            lecture_train_dataset.labels = lecture_train_dataset.labels + ["JUNK"] * n_extra_samples

        RecognitionHelper.set_dataset_mapping(lecture_train_dataset)
        RecognitionHelper.set_dataset_mapping(lecture_test_dataset, lecture_train_dataset.label_mapping,
                                              lecture_train_dataset.class_mapping)

        if junk_mode == 1:
            extra_train_dataset.labels = ["JUNK"] * n_extra_samples
            extra_test_dataset.labels = ["JUNK"] * n_testing_extra_samples
            RecognitionHelper.set_dataset_mapping(extra_train_dataset, lecture_train_dataset.label_mapping,
                                                  lecture_train_dataset.class_mapping)
            RecognitionHelper.set_dataset_mapping(extra_test_dataset, lecture_train_dataset.label_mapping,
                                                  lecture_train_dataset.class_mapping)

        task_end = time.time()
        total_lecture_overhead += task_end - task_start

        # train a classifier using only symbols that appear on math expressions from current lecture...
        task_start = time.time()
        print("-> Training ...", flush=True)

        if debug_load_classifier:
            print("DEBUG is on, reading classifier from file ....")
            in_file = open(classifier_filename, 'rb')
            classifier = pickle.load(in_file)
            in_file.close()
        else:
            # train classifier ...
            classifier = ClassifierTrainer.train_classifier(lecture_train_dataset, rec_config)

        if junk_mode == 0:
            threshold, t_data = RecognitionHelper.find_confidence_threshold(classifier, lecture_train_dataset,
                                                                            extra_train_dataset, min_threshold_recall)
            t_recall, t_precision, t_fmeasure = t_data
            print("-> Threshold found: " + str(threshold))
            print("---> Recall: {0:.4f}%".format(t_recall * 100.0))
            print("---> Precision: {0:.4f}%".format(t_precision * 100.0))
            print("---> F-Measure: {0:.4f}%".format(t_fmeasure * 100.0))
        else:
            threshold = 0.0

        task_end = time.time()
        total_lecture_training += task_end - task_start

        task_start = time.time()
        # evaluate on lecture training and testing sets according to Junk Mode ...
        if junk_mode == 0:
            # use threshold ... to reject
            train_info = RecognitionHelper.threshold_classifier_per_class_accuracy(classifier, lecture_train_dataset,
                                                                                   threshold)
            test_info = RecognitionHelper.threshold_classifier_per_class_accuracy(classifier, lecture_test_dataset,
                                                                                  threshold)
        else:
            # use regular pipeline as junk is a class
            train_info = EvalOps.classifier_per_class_accuracy(classifier, lecture_train_dataset)
            test_info = EvalOps.classifier_per_class_accuracy(classifier, lecture_test_dataset)

        train_acc, train_avg, train_std, _ = train_info
        test_acc, test_avg, test_std, errors = test_info

        print("-> Training Results")
        print("\tGlb. ACC\tPC. AVG\tPC. STD")
        print("\t{0:.4f}\t{1:.4f}\t{2:.4f}".format(train_acc, train_avg, train_std))

        print("-> Testing Results")
        print("\tGlb. ACC\tPC. AVG\tPC. STD")
        print("\t{0:.4f}\t{1:.4f}\t{2:.4f}".format(test_acc, test_avg, test_std))

        # check performance on isolated junk ...
        if junk_mode == 0:
            train_junk_acc = RecognitionHelper.threshold_classifier_junk_accuracy(classifier, extra_train_dataset,
                                                                                  threshold)
            test_junk_acc = RecognitionHelper.threshold_classifier_junk_accuracy(classifier, extra_test_dataset,
                                                                                 threshold)
        else:
            train_junk_acc, _, _, _ = EvalOps.classifier_per_class_accuracy(classifier, extra_train_dataset)
            test_junk_acc, _, _, _ = EvalOps.classifier_per_class_accuracy(classifier, extra_test_dataset)

        print("-> Junk")
        print("\tTrain\tTest")
        print("\t{0:.4f}\t{1:.4f}".format(train_junk_acc, test_junk_acc))

        # add for global averages ...
        all_train_acc.append(train_acc)
        all_train_avg.append(train_avg)
        all_train_std.append(train_std)

        all_test_acc.append(test_acc)
        all_test_avg.append(test_avg)
        all_test_std.append(test_std)

        all_junk_train.append(train_junk_acc)
        all_junk_test.append(test_junk_acc)

        task_end = time.time()
        total_lecture_testing += task_end - task_start

        task_start = time.time()
        print("-> Saving classifier")
        named_classifiers.append((lecture_title, rec_conf_filename, classifier_filename))

        out_file = open(classifier_filename, 'wb')
        pickle.dump(classifier, out_file, pickle.HIGHEST_PROTOCOL)
        if junk_mode == 0:
            pickle.dump(threshold, out_file, pickle.HIGHEST_PROTOCOL)
        out_file.close()
        task_end = time.time()
        total_lecture_saving += task_end - task_start

    print("\n\nSaving Named Classifiers Index to: " + named_classifiers_filename)
    task_start = time.time()
    with open(named_classifiers_filename, "w", encoding="utf8") as out_file:
        for rec_name, conf_filename, rec_path in named_classifiers:
            out_file.write("{0:s}\t{1:s}\t{2:s}\n".format(rec_name, conf_filename, rec_path))
    task_end = time.time()
    total_lecture_saving += task_end - task_start

    print("\n\t\tGlobal Averages")
    print("Training Results")
    print("\tGlb. ACC\tPC. AVG\tPC. STD")
    print("\t{0:.4f}\t{1:.4f}\t{2:.4f}".format(np.array(all_train_acc).mean(), np.array(all_train_avg).mean(),
                                               np.array(all_train_std).mean()))
    print("Testing Results")
    print("\tGlb. ACC\tPC. AVG\tPC. STD")
    print("\t{0:.4f}\t{1:.4f}\t{2:.4f}".format(np.array(all_test_acc).mean(), np.array(all_test_avg).mean(),
                                               np.array(all_test_std).mean()))

    print("Junk")
    print("\tTrain\tTest")
    print("\t{0:.4f}\t{1:.4f}".format(np.array(all_junk_train).mean(), np.array(all_junk_test).mean()))
    print("")

    end_time = time.time()
    print("Loading Time: " + TimeHelper.secondsToStr(end_loading_time - start_time))
    print("Total Lecture Pre-training Time: " + TimeHelper.secondsToStr(total_lecture_training))
    print("Total Lecture Training Time: " + TimeHelper.secondsToStr(total_lecture_training))
    print("Total Lecture Testing Time: " + TimeHelper.secondsToStr(total_lecture_testing))
    print("Total Saving Time: " + TimeHelper.secondsToStr(total_lecture_saving))
    print("Total Elapsed Time: " + TimeHelper.secondsToStr(end_time - start_time))

    print("Finished!")


if __name__ == "__main__":
    if sys.stdout.encoding != 'utf8':
      sys.stdout = codecs.getwriter('utf8')(sys.stdout.buffer, 'strict')
    if sys.stderr.encoding != 'utf8':
      sys.stderr = codecs.getwriter('utf8')(sys.stderr.buffer, 'strict')

    main()
