
import codecs
import os
import pickle
import sys
import time

from AM_CommonTools.util.time_helper import TimeHelper
from AccessMath.data.meta_data_DB import MetaDataDB
from AccessMath.indexing.lecture_class_mapping import LectureClassMapping
from AccessMath.indexing.tangents_helper import TangentS_Helper

from AM_CommonTools.configuration.configuration import Configuration as ConfigTanv
from TangentV.util.latex_helper import LaTeX_Helper

from TangentS.utility.control import Control

from recognizer.configuration.configuration import Configuration as ConfigRec


def main():
    # Load expressions per lecture
    if len(sys.argv) < 2:
        print("Usage:")
        print("")
        print(" python {0:s} database".format(sys.argv[0]))
        print("")
        print("Where")
        print("\tDatabase:\tPath to AccessMath database file (in XML format)")
        return

    start_time = time.time()

    database_filename = sys.argv[1]
    # rec_conf_filename = sys.argv[2]
    # tan_config_filename = sys.argv[3]

    start_task = time.time()
    # Load data ...
    # load database
    try:
        database = MetaDataDB.from_file(database_filename)
    except:
        print("Invalid database file")
        return

    rec_conf_filename = database.indexing.recognizer_config_filename
    tans_conf_filename = database.indexing.symbolic_mir_config_filename
    tanv_config_filename = database.indexing.image_mir_config_filename
    named_classifiers_filename = database.indexing.named_latex_classifiers_filename

    try:
        rec_config = ConfigRec.from_file(rec_conf_filename)
    except:
        print("Invalid Symbol Recognizer configuration")
        return

    base_classifier_name = rec_config.data["CLASSIFIER_DIR"] + "/" + rec_config.data["CLASSIFIER_NAME"] + "_LATEX"
    default_classifier_filename = base_classifier_name + ".dat"

    # tangent-s control
    try:
        tan_control = Control(tans_conf_filename) # control file name (after indexing)
    except:
        print("Invalid tangent-s control file ")
        return

    # load tangent-V config ...
    try:
        tanv_config = ConfigTanv.from_file(tanv_config_filename)
    except:
        print("Invalid Tangent-V configuration")
        return

    # ... load existing Full Mapping
    mapping_filename = rec_config.data["CLASSIFIER_DIR"] + "/per_lecture_classes.dat"
    print(mapping_filename)
    if os.path.isfile(mapping_filename):
        with open(mapping_filename, "rb") as in_file:
            full_mapping = pickle.load(in_file)
            assert isinstance(full_mapping, LectureClassMapping)
    else:
        print("Lecture-Mappings not found, must be created first")
        return

    # get the math expressions on each lecture ...
    exp_index = TangentS_Helper.load_index_expressions(tan_control)
    total_exp_count = 0
    for idx in exp_index:
        print("-> {0:s} expressions: {1:d}".format(full_mapping.lecture_mapping[idx].title, len(exp_index[idx])))
        total_exp_count += len(exp_index[idx])

    print("A total of {0:d} math expressions where found in {1:d} lectures".format(total_exp_count, len(exp_index)))

    latex_per_lecture = {}
    total_expressions, total_errors = 0, 0
    for idx in full_mapping.lecture_mapping:
        lecture_latex, lecture_errors = TangentS_Helper.load_document_latex_expressions(tan_control, idx)
        latex_per_lecture[idx] = lecture_latex

        total_errors += lecture_errors
        total_expressions += len(lecture_latex) - lecture_errors

    print("Total original expressions: " + str(total_expressions + total_errors))
    print("-> Total successfully extracted: " + str(total_expressions))
    print("-> Total failure cases: " + str(total_errors))

    end_task = time.time()
    total_time_loading = end_task - start_task

    # create the export paths ....
    # ... main ....
    main_export_dir = database.indexing.latex_image_path
    main_index_filename = main_export_dir + "/image_index.dat"
    if not os.path.isdir(main_export_dir):
        # create it ...
        os.mkdir(main_export_dir)

    # .... images .....
    export_image_dir = main_export_dir + "/images"
    export_image_index_filename = export_image_dir + "/index.txt"
    if not os.path.isdir(export_image_dir):
        os.mkdir(export_image_dir)
    # .... searches  .....
    export_search_dir = main_export_dir + "/searches"
    if not os.path.isdir(export_search_dir):
        os.mkdir(export_search_dir)
    # .... temporal  .....
    export_temporal_dir = main_export_dir + "/temporal"
    if not os.path.isdir(export_temporal_dir):
        os.mkdir(export_temporal_dir)

    default_classifier_info = (rec_conf_filename, default_classifier_filename)

    # export custom configuration
    # ... set new config parameters ....
    tanv_config.set("BINARY_IMAGES_DIRECTORY", export_image_dir)
    tanv_config.set("BINARY_IMAGES_METADATA", export_image_index_filename)
    tanv_config.set("INDEX_FILENAME", main_index_filename)
    tanv_config.set("SEARCH_OUTPUT_DIRECTORY", export_search_dir)
    tanv_config.set("CLASSIFIERS_CONFIG_DIRECTORY", ".")
    tanv_config.set("CLASSIFIERS_TRAINED_DIRECTORY", ".")
    tanv_config.set("TEMPORARY_CLASSIFIED_SYMBOLS", export_temporal_dir + "/classified_symbols_")
    tanv_config.set("SYMBOL_CLASSIFIER_DEFAULT", default_classifier_info)
    tanv_config.set("SYMBOL_CLASSIFIER_BY_NAME", named_classifiers_filename)
    tanv_config.set("METADATA_INDEX_BY", ["lecture"])
    # ... save ....
    tanv_config.save(database.indexing.latex_tanv_config_path)

    # Render Expressions to images
    start_task = time.time()
    print("\nRendering Expressions")

    count_errors = 0
    count_success = 0
    image_index_header = "instance_id\timg_location\tclassifier\tlecture\texpression_id\tLaTeX\n"
    image_index_lines = [image_index_header]
    for lecture_idx in full_mapping.lecture_mapping:
        title = full_mapping.lecture_mapping[lecture_idx].title

        print("Processing Lecture: " + title, flush=True)

        classifier_filename = base_classifier_name + "_" + title + ".dat"

        # dir for lecture images ...
        lecture_dir = export_image_dir + "/" + title
        if not os.path.isdir(lecture_dir):
            os.mkdir(lecture_dir)

        for idx, latex_str in enumerate(latex_per_lecture[lecture_idx]):
            image_suffix_name = "exp_" + str(idx) + ".png"
            image_filename = lecture_dir + "/" + image_suffix_name

            _, success = LaTeX_Helper.render(latex_str, image_filename)

            if not success:
                print("\nFailed Rendering")
                print("Lecture: " + str(title) + ", idx = " + str(idx))
                print(latex_str)
                count_errors += 1
                continue
            else:
                count_success += 1

                instance_id = title + ":" + str(idx)
                image_location = title + "/" + image_suffix_name
                classifier_name = title
                image_line = "\t".join([instance_id, image_location, classifier_name, title, str(idx), latex_str])
                image_index_lines.append(image_line + "\n")

    end_task = time.time()
    total_time_rendering = end_task - start_task

    print("Total images successfully rendered: " + str(count_success))
    print("Total images that failed to render: " + str(count_errors))

    # Save image index
    with open(export_image_index_filename, "w", encoding="utf8") as out_file:
        out_file.writelines(image_index_lines)

    end_time = time.time()

    print("Total Loading Time: " + TimeHelper.secondsToStr(total_time_loading))
    print("Total Rendering Time: " + TimeHelper.secondsToStr(total_time_rendering))
    print("Total Elapsed Time: " + TimeHelper.secondsToStr(end_time - start_time))
    print("Finished!")


if __name__ == '__main__':
    if sys.stdout.encoding != 'utf8':
      sys.stdout = codecs.getwriter('utf8')(sys.stdout.buffer, 'strict')
    if sys.stderr.encoding != 'utf8':
      sys.stderr = codecs.getwriter('utf8')(sys.stderr.buffer, 'strict')

    main()
