'''
Author : Parag Mali
Modified by: Kenny Davila

amtoimg_06_export_queries.py

Sample command to execute

> python3 amtoimg_06_export_queries.py test_data/databases/db_AccessMath2015.xml 99 0 0 amtoimg_06_ip amtoimg_06_out

This Script is part of the interface between AccessMath and Tangent:
 - script name: amtoimg_06_export_queries.py
   - parameters:
      - Database:               AccessMath Database file.
      - QuerySetID:             Name given to the query set.
      - Source Domain:          Source of queries (0 = HW, 1 = LaTeX)
      - Target Domain:          Target of queries (0 = HW, 1 = LaTeX)
      - Input Index:            Path to a CSV file containing location and source information for input queries
      - Output Directory:       Path where query set will be stored.

  - If the wrong or no parameters are provided, the script is going to print
    an easy to understand help, not just a list of parameters, but also prints what each
    of them means and the values they might take.

  - The input index file in CSV format will be expected to have a list of entries, one
    per query, using the following format:
      query-id, path-to-image, binary*, source**

    * Binary is 1 if the given image is already binary, or 0 otherwise.
   ** Note that the source attribute is optional.
   By default, query images without a source will be associated to the
   "default" source which will use a default classifier for the images
   instead of lecture-specific classifier. Note that valid source names should
   consistent with lecture titles on the current database. Please validate each
    source name exists in the given input database.

 - The script must produce a valid Tangent-V XML query set file. This file will be stored
    in [output Directory]/[QuerySetID].xml.

 - Images references by each query will be copied to a sub-directory at
  [OutputDirectory]/[QuerySetID].  Each input image copied to the output subdirectory
  is named "[OutputDirectory] /[QuerySetID]/[Query-ID].png"

  - When copying images, the script checks the binary attribute, and for non-binary images
  (binary =0), our MLBinarizer is applied first to make them binary during the export
  process. This will make super easy to take snapshots of videos, cut the query region, add
  this flag, and then the query export script will convert them to binary images compatible
  with Tangent-V automatically. Other general image sources like pictures of handwritten math
  expressions could be welcome here in the near future.

 - XML file format:
<AccessMathQueryFile>
     <QuerySetInfo>
                <Id>[Query Set ID]</Id>
		<Name> [Query Set ID] </Name>
		<Description>[SOURCE] Image test query set containing single image queries only (Type 0 = Single-Image)</Description>
		<NamedClassifiers>[PATH TO NAMED CLASSIFIERS FILE from AccessMathDB based on SOURCE]</NamedClassifiers>
     </QuerySetInfo>
     <QuerySet>
               [ LIST OF QUERIES ]
     </QuerySet>
</AccessMathQueryFile>

  - The list of queries will contain a list of <Query> entries, one per query, with some different options set depending on the target domain.
       <Query>
			<Id>[Query-ID]</Id>
			<Type>Single-Image</Type>
			<Groups>
				[TARGET - DOMAIN DEPENDENT]
			</Groups>
			<Images>
				<Image>
					<File>[QuerySetID]/[Query-ID].png</File>
					<Color>BIN</Color>
					<Type>HW</Type>
					<Source>[Source]</Source>
				</Image>
			</Images>
		</Query>

     -> For Queries targeted to retrieve Handwriting, the groups tag must be produced like this:
			<Groups>
				<Mode>Sequence</Mode>
				<GroupBy>lecture</GroupBy>
				<SequenceBy>frame_idx</SequenceBy>
			</Groups>

     -> For Queries targeted to retrieve LaTeX, the groups tag must be produced like this:
			<Groups>
				<Mode>Attribute</Mode>
				<GroupBy>lecture</GroupBy>
				<GroupBy>LaTeX</GroupBy>
			</Groups>

  - The produced XML file is also human readable which means with the proper end of lines and indentation
   levels. BeautifulSoup is not used, since it is NOT case sensitive, and it will ignore the casing
   of the XML tags. Note that the tag name processing of standard XML Parsers IS case sensitive.

'''

import sys
import pickle
import codecs

from AccessMath.preprocessing.config.parameters import Parameters
from AccessMath.preprocessing.content.MLBinarizer import MLBinarizer
from AccessMath.data.meta_data_DB import MetaDataDB

from TangentV.util.query_exporter import QueryExporter


class AM_QueryValidator:
    def __init__(self, database_filename):
        try:
            self.database = MetaDataDB.from_file(database_filename)
        except:
            raise Exception("Invalid AccessMath database file")

    def validate_source(self, source):
        if source.lower() == "default":
            # allow usage of default classifier
            return True

        # find lecture
        current_lecture = self.database.get_lecture(source)

        # valid if lecture is found
        return current_lecture is not None


def load_binarizer():
    # load classifier from file ...
    in_file = open(Parameters.MLBin_classifier_file, "rb")
    classifier = pickle.load(in_file)
    in_file.close()

    ml_binarizer = MLBinarizer(classifier, Parameters.MLBin_patch_size, Parameters.MLBin_sigma_color,
                               Parameters.MLBin_sigma_space, Parameters.MLBin_median_blur_k,
                               Parameters.MLBin_dark_background)

    return ml_binarizer

def main():
    # check command line arguments
    if len(sys.argv) < 7:
        print("Usage:")
        print("")
        print(" python {0:s} Database Query_Set_ID Source_Domain Target_Domain Query_Index Output_directory".format(sys.argv[0]))
        print("")
        print("Where")
        print("\tDatabase:\tPath to AccessMath database file (in XML format)")
        print("\tQuery_Set_ID:\tName given to query dataset")
        print("\tSource_Domain:\tSource of queries (0=HW, 1=LaTeX)")
        print("\tTarget_Domain:\tTarget of queries (0=HW, 1=LaTeX)")
        print("\tQuery_Index:\tPath to a CSV file containing location and source information for input queries")
        print("\tOutput_directory:\tPath where query set will be stored")

        return

    # all parameters
    database_filename = sys.argv[1]
    query_set_id = sys.argv[2]
    source_domain = sys.argv[3]
    target_domain = sys.argv[4]
    index_filename = sys.argv[5]
    output_dir = sys.argv[6]

    # Prepare exporter helpers .
    # 1) Source validator
    validator = AM_QueryValidator(database_filename)
    # 2) image binarizer
    binarizer = load_binarizer()

    # based on source domain
    if source_domain == "1":
        # LaTeX
        query_type = "LATEX"
        named_classifier = validator.database.indexing.named_latex_classifiers_filename
    else:
        # Handwritten
        query_type = "HW"
        named_classifier = validator.database.indexing.named_hw_classifiers_filename

    # based on target domain
    if target_domain == "1":
        # To LaTeX
        group_info = {
            "mode": "Attribute",
            "group_by": ["lecture", "LaTeX"],
        }
    else:
        # To Handwriting ...
        group_info = {
            "mode": "Sequence",
            "group_by": ["lecture"],
            "sequence_by": ["frame_idx"],
        }

    query_color = "BIN"

    # prepare exporter ...
    query_exporter = QueryExporter(binarizer.hysteresis_binarize, validator)
    # load query metadata
    query_index = query_exporter.load_query_index(index_filename)
    # export!
    query_exporter.export(query_set_id, query_index, query_type, query_color, named_classifier, group_info, output_dir)

if __name__ == '__main__':
    if sys.stdout.encoding != 'utf8':
        sys.stdout = codecs.getwriter('utf8')(sys.stdout.buffer, 'strict')
    if sys.stderr.encoding != 'utf8':
        sys.stderr = codecs.getwriter('utf8')(sys.stderr.buffer, 'strict')

    main()