
#=====================================================================
#  This class represents a traces, and also performs all the operations
#  relative to it
#
#  Created by:
#      - Kenny Davila (May, 2016)
#=====================================================================
import math
import sys
import numpy as np
import pickle

from concurrent.futures import ProcessPoolExecutor
from recognizer.data.code_book import CodeBook
from recognizer.features.crossings_feature_extractor import CrossingsFeatureExtractor
from recognizer.features.histograms_feature_extractor import HistogramFeatureExtractor
from recognizer.features.trace_feature_extractor import TraceFeatureExtractor
from recognizer.features.code_book_feature_extractor import CodeBookFeatureExtractor
from recognizer.features.temporal_feature_extractor import TemporalFeatureExtractor

class FeatureExtractor:
    def __init__(self, configuration):
        self.config = configuration

        if self.config.get_bool("FEATURES_USE_TEMPORAL_FEATURES"):
            self.temporal_feature_extractor = TemporalFeatureExtractor(configuration)
        else:
            self.temporal_feature_extractor = None

        if self.config.get_bool("FEATURES_USE_CODE_BOOK_HISTOGRAMS"):
            # Create Code-book feature extractor
            self.cb_feature_extractor = CodeBookFeatureExtractor.create_from_config(configuration)
            self.code_book = CodeBook.load_from_file(configuration)
        else:
            self.cb_feature_extractor = None
            self.code_book = None

    def getCrossingsFeatures(self, symbol):
        features = []
        n_crossings = self.config.get_int("CROSSINGS_NUMBER", 5)
        n_subcrossings = self.config.get_int("CROSSINGS_SUBCROSSINGS", 5)
        use_overlapping = self.config.get_bool("CROSSINGS_OVERLAPPING", False)
        sub_weights = self.config.get_int("SUBCROSSINGS_WEIGHTS", 0)
        sub_sigma = self.config.get_float("SUBCROSSINGS_GAUSS_SIGMA", 1.0)

        # compute crossings ...
        all_crossings = CrossingsFeatureExtractor.getHorizontalVerticalCrossings(symbol, n_crossings, n_subcrossings,
                                                                                 sub_weights, sub_sigma, use_overlapping)

        if self.config.get_bool("CROSSINGS_USE_COUNT"):
            features += all_crossings["counts"]
        if self.config.get_bool("CROSSINGS_USE_MIN"):
            features += all_crossings["mins"]
        if self.config.get_bool("CROSSINGS_USE_MAX"):
            features += all_crossings["maxs"]

        return features

    def getDiagonalCrossingsFeatures(self, symbol):
        features = []
        n_crossings = self.config.get_int("DIAGONAL_CROSSINGS_NUMBER", 5)
        n_subcrossings = self.config.get_int("DIAGONAL_CROSSINGS_SUBCROSSINGS", 5)
        use_overlapping = self.config.get_bool("DIAGONAL_CROSSINGS_OVERLAPPING", False)
        sub_weights = self.config.get_int("DIAGONAL_SUBCROSSINGS_WEIGHTS", 0)
        sub_sigma = self.config.get_float("DIAGONAL_SUBCROSSINGS_GAUSS_SIGMA", 1.0)

        all_crossings = CrossingsFeatureExtractor.getDiagonalCrossings(symbol, n_crossings, n_subcrossings,
                                                                       sub_weights, sub_sigma, use_overlapping)

        if self.config.get_bool("DIAGONAL_CROSSINGS_USE_COUNT"):
            features += all_crossings["counts"]
        if self.config.get_bool("DIAGONAL_CROSSINGS_USE_MIN"):
            features += all_crossings["mins"]
        if self.config.get_bool("DIAGONAL_CROSSINGS_USE_MAX"):
            features += all_crossings["maxs"]

        return features

    def getAngularCrossingFeatures(self, symbol):
        n_angular = self.config.get_int("ANGULAR_CROSSINGS_NUMBER", 4)
        n_subcrossings = self.config.get_int("ANGULAR_CROSSINGS_SUBCROSSINGS", 1)
        sub_weights = self.config.get_int("ANGULAR_SUBCROSSINGS_WEIGHTS", 0)
        sub_sigma = self.config.get_float("ANGULAR_SUBCROSSINGS_GAUSS_SIGMA", 1.0)

        all_crossings = CrossingsFeatureExtractor.getAngularCrossings(symbol, n_angular, n_subcrossings,
                                                                      sub_weights, sub_sigma)

        features = []
        if self.config.get_bool("ANGULAR_CROSSINGS_USE_COUNT"):
            features += all_crossings["counts"]
        if self.config.get_bool("ANGULAR_CROSSINGS_USE_MIN"):
            features += all_crossings["mins"]
        if self.config.get_bool("ANGULAR_CROSSINGS_USE_MAX"):
            features += all_crossings["maxs"]

        return features

    def getZoningCrossingsFeatures(self, symbol):
        grids = self.config.get("ZONING_CROSSINGS_GRIDS")

        n_hor_lines = self.config.get_int("ZONING_CROSSINGS_HOR_LINES", -1)
        n_ver_lines = self.config.get_int("ZONING_CROSSINGS_VER_LINES", -1)
        n_dia_lines = self.config.get_int("ZONING_CROSSINGS_DIA_LINES", -1)
        weight_function = self.config.get_int("ZONING_CROSSINGS_WEIGHTS", -1)
        weight_sigma = self.config.get_float("ZONING_CROSSINGS_GAUSS_SIGMA", 1.0)

        if n_hor_lines < 1:
            raise Exception("Invalid value for <ZONING_CROSSINGS_HOR_LINES>")

        if n_ver_lines < 0:
            raise Exception("Invalid value for <ZONING_CROSSINGS_VER_LINES>")
        elif n_ver_lines == 0:
            # use the same number of vertical lines as horizontal lines
            n_ver_lines = n_hor_lines

        if n_dia_lines < 0:
            raise Exception("Invalid value for <ZONING_CROSSINGS_DIA_LINES>")
        elif n_dia_lines == 0:
            n_dia_lines = n_hor_lines

        if weight_function < 0:
            raise Exception("Invalid value for <ZONING_CROSSINGS_WEIGHTS>")

        return CrossingsFeatureExtractor.getZoningCrossings(symbol, n_hor_lines, n_ver_lines, n_dia_lines,
                                                            weight_function, weight_sigma, grids)


    def get2DHistograms(self, symbol):
        grids = self.config.get("2D_HISTOGRAMS_GRIDS")

        return HistogramFeatureExtractor.computeSymbol2DHistogram(symbol, grids)

    def getOrientationHistograms(self, symbol):
        grids = self.config.get("ORIENTATION_HISTOGRAMS_GRIDS")

        return HistogramFeatureExtractor.computeSymbolOrientationHistogram(symbol, grids)

    def getTraceDescriptionFeatures(self, symbol):
        features = []
        if self.config.get_bool("TRACES_DESCRIPTION_USE_COUNT"):
            features += TraceFeatureExtractor.getTracesCount(symbol)

        if self.config.get_bool("TRACES_DESCRIPTION_USE_ASPECT_RATIO"):
            features += TraceFeatureExtractor.getAspectRatio(symbol)

        # Angular Change
        use_angular_change_total = self.config.get_bool("TRACES_DESCRIPTION_USE_TOTAL_ANGULAR_CHANGE")
        use_angular_change_average = self.config.get_bool("TRACES_DESCRIPTION_USE_AVERAGE_ANGULAR_CHANGE")
        if use_angular_change_total or use_angular_change_average:
            ang_change_total, ang_change_avg = TraceFeatureExtractor.getAngularChange(symbol)

            if use_angular_change_total:
                features.append(ang_change_total)

            if use_angular_change_average:
                features.append(ang_change_avg)

        # Total Line Length (After smoothing and normalization)
        use_line_length_total = self.config.get_bool("TRACES_DESCRIPTION_USE_TOTAL_LINE_LENGTH")
        use_line_length_average = self.config.get_bool("TRACES_DESCRIPTION_USE_AVERAGE_LINE_LENGTH")
        if use_line_length_total or use_line_length_average:
            line_length_total, line_length_avg = TraceFeatureExtractor.getLineLength(symbol)

            if use_line_length_total:
                features.append(line_length_total)

            if use_line_length_average:
                features.append(line_length_avg)

        # Count of Sharp Points
        use_sharp_points_total = self.config.get_bool("TRACES_DESCRIPTION_USE_TOTAL_SHARP_POINTS")
        use_sharp_points_average = self.config.get_bool("TRACES_DESCRIPTION_USE_AVERAGE_SHARP_POINTS")
        if use_sharp_points_total or use_sharp_points_average:
            sharp_points_total, sharp_points_avg = TraceFeatureExtractor.getNumberOfSharpPoints(symbol)

            if use_sharp_points_total:
                features.append(sharp_points_total)

            if use_sharp_points_average:
                features.append(sharp_points_avg)

        use_var_x = self.config.get_bool("TRACES_DESCRIPTION_USE_VARIANCE_X")
        use_var_y = self.config.get_bool("TRACES_DESCRIPTION_USE_VARIANCE_Y")
        use_covar_xy = self.config.get_bool("TRACES_DESCRIPTION_USE_COVARIANCE_XY")
        if use_var_x or use_var_y or use_covar_xy:
            var_x, var_y, cov_xy = TraceFeatureExtractor.getVarianceFeatures(symbol)

            if use_var_x:
                features.append(var_x)

            if use_var_y:
                features.append(var_y)

            if use_covar_xy:
                features.append(cov_xy)

        use_mid_point_var_x = self.config.get_bool("TRACES_DESCRIPTION_USE_MID_POINT_VARIANCE_X")
        use_mid_point_var_y = self.config.get_bool("TRACES_DESCRIPTION_USE_MID_POINT_VARIANCE_Y")
        use_mid_point_covar_xy = self.config.get_bool("TRACES_DESCRIPTION_USE_MID_POINT_COVARIANCE_XY")
        if use_mid_point_var_x or use_mid_point_var_y or use_mid_point_covar_xy:
            mp_var_x, mp_var_y, mp_cov_xy = TraceFeatureExtractor.getMidPointVarianceFeatures(symbol)

            if use_mid_point_var_x:
                features.append(mp_var_x)

            if use_mid_point_var_y:
                features.append(mp_var_y)

            if use_mid_point_covar_xy:
                features.append(mp_cov_xy)

        if self.config.get_bool("TRACES_DESCRIPTION_TOTAL_ORIGINAL_POINTS"):
            features.append(TraceFeatureExtractor.getTotalOriginalPoints(symbol))

        return features

    def getCodeBookFeatures(self, symbol):
        all_features, all_boxes = self.cb_feature_extractor.getFeatures(symbol)

        features = self.cb_feature_extractor.getSymbolEncoding(all_boxes, all_features, self.code_book)

        return features.tolist()

    def getFeatures(self, symbol):
        features = []

        # 1) first, crossings...
        if self.config.get_bool("FEATURES_USE_CROSSINGS"):
            features += self.getCrossingsFeatures(symbol)

        # 1.1) Diagonal crossings...
        if self.config.get_bool("FEATURES_USE_DIAGONAL_CROSSINGS"):
            features += self.getDiagonalCrossingsFeatures(symbol)

        # 1.5) Angular crossings ...
        if self.config.get_bool("FEATURES_USE_ANGULAR_CROSSINGS"):
            features += self.getAngularCrossingFeatures(symbol)

        # 1.9) Zoning Crossings ...
        if self.config.get_bool("FEATURES_USE_ZONING_CROSSINGS"):
            features += self.getZoningCrossingsFeatures(symbol)

        # 2) 2D Histograms ...
        if self.config.get_bool("FEATURES_USE_2D_HISTOGRAMS"):
            features += self.get2DHistograms(symbol)

        # 3) Orientation Histograms ...
        if self.config.get_bool("FEATURES_USE_ORIENTATION_HISTOGRAMS"):
            features += self.getOrientationHistograms(symbol)

        # 4) General Features
        # 4.1) Number of traces ...
        if self.config.get_bool("FEATURES_USE_TRACES_DESCRIPTION"):
            features += self.getTraceDescriptionFeatures(symbol)

        # 5) Code-Book based features
        if self.config.get_bool("FEATURES_USE_CODE_BOOK_HISTOGRAMS"):
            features += self.getCodeBookFeatures(symbol)

        # 6) features based on temporal info
        if self.config.get_bool("FEATURES_USE_TEMPORAL_FEATURES"):
            features += self.temporal_feature_extractor.getFeatures(symbol)

        if len(features) == 0:
            # for debugging only, will prefer majority class
            features.append(0.0)

        return features

    @staticmethod
    def getSymbolsFeatures(symbols, config):
        feature_extractor = FeatureExtractor(config)

        symbol_features = []
        for math_symbol in symbols:
            symbol_features.append(feature_extractor.getFeatures(math_symbol))

        return symbol_features

    @staticmethod
    def getSymbolsFeaturesParallel(symbols, config, n_workers):
        feature_extractor = FeatureExtractor(config)

        symbol_features = []
        with ProcessPoolExecutor(max_workers=n_workers) as executor:
            for current_features in executor.map(feature_extractor.getFeatures, symbols):
                symbol_features.append(current_features)

        return symbol_features