import os
import sys
import fnmatch
import xml.etree.ElementTree as ET

from concurrent.futures import ProcessPoolExecutor
from recognizer.data.feature_dataset import FeatureDataset
from recognizer.symbols.math_symbol_trace import MathSymbolTrace
from recognizer.symbols.math_symbol import MathSymbol
from recognizer.preprocessing.trace_smoother import TraceSmoother

class INKMLLoader:
    INKML_NAMESPACE = '{http://www.w3.org/2003/InkML}'

    def instance_process_INKML_file(self, data):
        return INKMLLoader.process_INKML_file(data)

    @staticmethod
    def path_INKML_filenames(path):
        complete_list = os.listdir(path)
        filtered_list = []
        for local_name in complete_list:
            full_name = path + "/" + local_name
            if os.path.isdir(full_name):
                child_filenames = INKMLLoader.path_INKML_filenames(full_name)
                for child_filename in child_filenames:
                    filtered_list.append(local_name + "/" + child_filename)
            else:
                if fnmatch.fnmatch(local_name, '*.inkml'):
                    filtered_list.append(local_name)

        return filtered_list

    @staticmethod
    def load_label_replacement(filename):
        # load file
        replacement_labels_file = open(filename, "r")
        lines = replacement_labels_file.readlines()
        replacement_labels_file.close()

        replacement_labels = {}
        for line in lines:
            parts = line.strip().split("\t")
            if len(parts) == 2:
                replacement_labels[parts[0]] = parts[1]

        return replacement_labels

    @staticmethod
    def load_INKML_annotations(root):
        all_annotations = {}
        for annotation in root.findall(INKMLLoader.INKML_NAMESPACE + 'annotation'):
            try:
                key = annotation.attrib["type"]
                value = annotation.text

                all_annotations[key] = value
            except:
                # ignore annotation
                continue

        return all_annotations

    @staticmethod
    def load_INKML_traces(root, debug_raw_prefix=None, debug_added_prefix=None, debug_smooth_prefix=None):

        # extract all the traces first...
        trace_objects = {}
        for trace in root.findall(INKMLLoader.INKML_NAMESPACE + 'trace'):
            # load
            object_trace = MathSymbolTrace.from_INKML_trace(trace)

            # apply general trace pre processing...
            TraceSmoother.applyPreprocessing(object_trace, debug_raw_prefix, debug_added_prefix, debug_smooth_prefix)

            # add to the dictionary...
            trace_objects[object_trace.id] = object_trace

        return trace_objects

    @staticmethod
    def load_INKML_symbols(root, trace_objects, ground_truth_available, active_classes=None, debug_normalization_prefix=None):
        # put all the traces together with their corresponding symbols...
        # first, find the root of the trace groups...

        groups_root = root.find(INKMLLoader.INKML_NAMESPACE + 'traceGroup')
        trace_groups = groups_root.findall(INKMLLoader.INKML_NAMESPACE + 'traceGroup')

        symbols = []
        avg_width = 0.0
        avg_height = 0.0
        for group in trace_groups:
            if ground_truth_available:
                # search for class label...
                symbol_class = group.find(INKMLLoader.INKML_NAMESPACE + 'annotation').text

                if symbol_class is None:
                    symbol_class = "junk"

                if active_classes is not None and not symbol_class in active_classes:
                    # skip this symbol since it does not belong to the current set of classes
                    continue

            else:
                # unknown class and id label
                symbol_class = '{Unknown}'

            # search for id attribute...
            symbol_id = 0
            for id_att_name in group.attrib:
                if id_att_name[-2:] == "id":
                    try:
                        symbol_id = int(group.attrib[id_att_name])
                    except:
                        # could not convert to int, try spliting...
                        symbol_id = int(group.attrib[id_att_name].split(":")[0])

            # link with corresponding traces...
            group_traces = group.findall(INKMLLoader.INKML_NAMESPACE + 'traceView')
            symbol_list = []
            for trace in group_traces:
                object_trace = trace_objects[int(trace.attrib["traceDataRef"])]
                symbol_list.append(object_trace)

            # create the math symbol...
            new_symbol = MathSymbol(symbol_id, symbol_list, symbol_class)

            # capture statistics of relative size....
            symMinX, symMaxX, symMinY, symMaxY = new_symbol.original_box
            # ...add...
            avg_width  += (symMaxX - symMinX)
            avg_height += (symMaxY - symMinY)

            # now normalize size and locations for traces in current symbol
            new_symbol.normalize()

            if debug_normalization_prefix is not None:
                # output data after relocated
                for trace in group_traces:
                    object_trace = trace_objects[int(trace.attrib["traceDataRef"])]

                    file = open(debug_normalization_prefix + str(object_trace.id) + '.txt', 'w')
                    file.write(str(object_trace))
                    file.close()

            symbols.append(new_symbol)

        if len(trace_groups) > 0:
            avg_width /= len(trace_groups)
            avg_height /= len(trace_groups)

            for s in symbols:
                s.setSizeRatio(avg_width, avg_height)

        return symbols

    @staticmethod
    def load_INKML_file(filename, ground_truth_available, active_classes=None):
        #first load the tree...
        tree = ET.parse(filename)
        root = tree.getroot()

        traces_objects = INKMLLoader.load_INKML_traces(root)
        symbols = INKMLLoader.load_INKML_symbols(root, traces_objects, ground_truth_available, active_classes)

        return symbols

    @staticmethod
    def process_INKML_file(data):
        file_path, ground_truth_available, feature_extractor, label_replacement, active_classes = data

        try:
            symbols = INKMLLoader.load_INKML_file(file_path, ground_truth_available, active_classes)
            success = (True, None)
        except Exception as e:
            success = (False, e)
            symbols = []

        symbol_data = []
        label_data = []
        source_data = []
        for new_symbol in symbols:
            if feature_extractor is not None:
                if isinstance(feature_extractor, list):
                    # multiple feature extractors
                    current_data = [new_symbol]
                    for extractor in feature_extractor:
                        current_data.append(extractor.getFeatures(new_symbol))

                    symbol_data.append(current_data)
                else:
                    # one single feature extractor
                    symbol_data.append(feature_extractor.getFeatures(new_symbol))
            else:
                symbol_data.append(new_symbol)

            # ... correct labels ...
            if new_symbol.truth in label_replacement:
                new_symbol.truth = label_replacement[new_symbol.truth]

            symbol_label = new_symbol.truth

            label_data.append(symbol_label)

            # the source of current symbol will stored
            source_data.append((file_path, new_symbol.id))

        return file_path, success, (symbol_data, label_data, source_data)

    @staticmethod
    def load_INKML_directory(path, ground_truth_available, feature_extractor, label_replacement, save_sources,
                             label_mapping=None, class_mapping=None, verbose=True, workers=10, active_classes=None):

        complete_list = INKMLLoader.path_INKML_filenames(path)

        raw_data, raw_labels, valid_files, error_files = [], [], [], []
        sources = [] if save_sources else None

        # create parameter array (for parallel execution)
        dir_data = []
        for i in range(len(complete_list)):
            file_name = complete_list[i]
            file_path = path + '/' + file_name

            file_data = (file_path, ground_truth_available, feature_extractor, label_replacement, active_classes)

            dir_data.append(file_data)

        # read in parallel every file in the path specified...
        empty_string = " " * 80
        with ProcessPoolExecutor(max_workers=workers) as executor:
            if sys.version_info.major == 3 and sys.version_info.minor >= 5:
                mapper = executor.map
                loading_function = INKMLLoader.process_INKML_file
            else:
                # avoid issues in versions older than python 3.5
                #mapper = map
                mapper = executor.map
                tempo_instance = INKMLLoader()
                loading_function = tempo_instance.instance_process_INKML_file
            
            for i, (file_path, success, file_data) in enumerate(mapper(loading_function, dir_data)):
                if verbose:
                    advance = float(i) / len(complete_list)
                    print(empty_string, end="\r")
                    print(("Processing => {:.2%} => " + file_path).format(advance), end="\r")

                symbol_data, label_data, source_data = file_data

                loading_success, error_desc = success
                if not loading_success:
                    if verbose:
                        print("Failed processing: " + file_path)
                        print(error_desc)
                    error_files.append(file_path)
                else:
                    valid_files.append(file_path)

                raw_data += symbol_data
                raw_labels += label_data

                if save_sources:
                    sources += source_data

        if feature_extractor is not None and not isinstance(feature_extractor, list):
            # exactly one feature extractor was specified, create dataset with extracted features
            dataset = FeatureDataset.from_INKML_data(raw_data, raw_labels, label_mapping, class_mapping, sources)
        else:
            # No feature extractor specified, return raw symbol data or
            # multiple feature extractor were specified, return array of feature sets extracted
            dataset = (raw_data, raw_labels, sources)

        return dataset, valid_files, error_files
